/*!
    \file   df_api.h
    \author Bernd Ottmann
    \date   21.02.2003
    \brief  Hauptinclude mit den allgemeinen Funktionen und Einbindung aller Untermodule.
*/
#ifndef DF_API_H
#define DF_API_H

#include "df_defines.h"
#include "df_connection.h"
#include "df_files.h"
#include "df_lists.h"
#include "df_log.h"
#include "df_setup_records.h"
#include "df_setup_lists.h"
#include "df_access_control_lists_v1.h"
#include "df_access_control.h"
#include "df_access_control_lists.h"
#include "df_system_records.h"
#include "df_fingerprint.h"
#include "df_active_mode.h"

/*! 
    \mainpage Dokumentation zur Datafox Kommunikationsbibliothek<br />DFComDLL / DFCom_x64
    \tableofcontents

    \section Introduction Einleitung
    Mit Hilfe der Kommunikationsbibliothek "DFComDLL, DFCom_x64" können die Datafox-Geräte Konfiguriert werden. Erzeugte Datensätze können von den Geräten gelesen und Listendaten auf
    die Geräte geschrieben werden. All diese Aufgaben können über unterschiedliche Kommunikationswege wie RS232, USB, bzw. RS485 sowie LAN/WLAN geschehen.

    Auch die Datafox Konfigurationssoftware DatafoxStudioIV verwendet für die unterschiedlichen Aufgaben und den Datenaustausch diese Kommunikationsbibliothek.
    Durch die Abwärtskompatibilithät werden durch die Kommunikationsroutinen das AEIII+, die MasterIV-Serie und EVO-Line Geräte unterstützt. Somit unterstützen Sie durch den Einsatz gleich eine
    ganze Menge an unterschiedlichster Hardware von Datafox.

	![Alt Einbindung](../images/overview_de.png "Einbindungsmethoden zur Datenübertragung MasterIV-Serie.")
	
    \subsection Overview_DatafoxStudioIV Datatafox Studio
    Einstellung der Geräte für die jeweilige Aufgabe der Datenerfassung. Die Konfigurationssoftware DatafoxStudioIV stellt die Anwendung zur Einrichtung der Datafox-Geräte dar.

    \subsection Overview_DFComDLL Kommunikationsbibliothek  (Direkte DLL Anbindung)
    Über die Kommunikationsbibliothek ist eine direkte Kommunikation zwischen der Anwendungssoftware und Gerät möglich.
    Dies ist damit die beste Art der Anbindung für Softwarehersteller (OEMs). Die Kommunikation über die Kommunikationsbibliothek ist bidirektional und kann sowohl online als auch offline erfolgen.

    \subsection Overview_Http HTTP (Webserver)
    Die Datafox-Geräte können Datensätze über das Http-Protokoll an einen ggf. Webserver senden. Über die Rückantwort können unterschiedliche zusätzliche Aktionen angestoßen werden,
    z. B. stellen der Uhrzeit oder anzeigen einer Nachricht auf der Geräteanzeige.
    
    Auf der Produkt-DVD stellt Datafox Ihnen ein PHP-Muster-Script zur Datenentgegennahme und Quittierung bereit.

    \subsection Overview_Talk Datafox-Talk
    Ein bereits entwickeltes Schnittstellenprogramm mit entsprechendem Funktionsumfang zum Datenaustausch über Dateien oder Datenbank.
    Diese Lösung ist ideal für Endkunden oder Händler, die Datafox-Produkte an z.B. Warenwirtschaft, ERP, ... anbinden wollen. 


    \section VersionChanges Versionsänderungen
    <table>
    <tr><th>Datum</th><th>Version</th><th>Änderungen</th></tr>
    <tr><td colspan="2">Version 4.3</td><td>Unterstützung der EVO-Line Serie.</td></tr>
    <tr><td>12.12.13</td><td>04.03.00</td><td>Unterstützung der Gerätetypen EVO 2.8, EVO-Line 4.2, IPC-Extension. Unterstützung der Firmware-Version 04.03.00.</td></tr>
    <tr><td>06.01.14</td><td>04.03.01</td><td>Unterstützung der Firmware-Version 04.03.01.</td></tr>
    <tr><td>20.03.14</td><td>04.03.02</td><td></td>Unterstützung der Firmware-Version 04.03.02. </tr>
    <tr><td>15.10.14</td><td>04.03.03</td><td></td>Unterstützung der Firmware-Version 04.03.03.</tr>
    <tr><td>19.01.15</td><td>04.03.04</td><td>Erweiterung um die Funktionen DFCAccessControlIdentification, DFCAccessControlOnlineAction zur Unterstützung mehrerer Zutrittsbusse.</td></tr>      
    </table>
*/

/*!
    \page WhatIsDLL Funktionsweise der DLL
    Die DLL ist die Schnittstelle zwischen einer Softwareanwendung und den AEIII+ / MasterIV - Geräten. Alle die Geräte betreffenden Aufgaben werden von der Softwareanwendung an die DLL übergeben, 
    dort verarbeitet und gegebenenfalls an das anzusprechende Gerät weitergeleitet.

    \section RecordHandling Behandlung von Datensätzen
    Um Datensätze aus den Geräten zu lesen muss zuerst die jeweilige Kommunikationsschnittstelle mit einer der Funktionen DFCComOpenSerial, DFCComOpenSocket (AEIII+) oder DFCComOpenIV (MasterIV) erfolgreich 
    initialisiert werden.

    \section UsedTechnicWithAEIII Verwendende Technik bei AEIII+
    Danach wird mit der Methode DFCComCollectData das Gerät mit der angegebenen Busnummer bekannt gegeben und der erste Datensatz (oder alle Daten, je nach Firmwareversion) aus dem Gerät gelesen.
    Die Funktion DFCComGetDatensatz bestätigt diesen zuvor durch DFCComCollectData ausgelesenen Datensatz, übergibt die Daten in den zur verfügung zu stellenden Datenpuffer und liest den ggf. nächsten 
    Datensatz aus dem Gerätespeicher aus. Dieser wird dann bei dem nächsten Aufruf von DFCComGetDatensatz quittiert und in den Datenpuffer übergeben. Auf diese Weise werden durch folgeaufrufe der Funktion 
    DFCComGetDatensatz einzelne Datensätze aus dem Gerät an die Anwendung übergeben.

    Zu verwendende Technik bei MasterIV
    Danach wird mit der Methode DFCReadRecord das Gerät angesprochen und der erste Datensatz aus dem Gerät gelesen. Nach erfolgreicher Verarbeitung des Datensatzes durch Ihre Anwendung muss mit der Funktion 
    DFCQuitRecord der zuvor gelesenen Datensatz quittiert werden. Erst nach erfolgreicher Ausführung der quittierung wird durch die Firmware auf den nächsten Datensatz gewechselt.
    Durch erneuten Aufruf der Funktion bestätigt diesen zuvor durch DFCComCollectData ausgelesenen Datensatz, übergibt die Daten in den zur verfügung zu stellenden Datenpuffer und liest den ggf. nächsten 
    Datensatz aus dem Gerätespeicher aus. Dieser wird dann bei dem nächsten Aufruf von DFCComGetDatensatz quittiert und in den Datenpuffer übergeben. Auf diese Weise werden durch folgeaufrufe der Funktion 
    DFCComGetDatensatz einzelne Datensätze aus dem Gerät an die Anwendung übergeben.

    Nach Ausführung aller benötigten Kommunikationsbefehle, muss mit der Funktion DFCComClose die Kommunikationsschnittstelle wieder geschlossen werden.
    Es kann während der Kommunikation kein Datensatz verloren gehen, bei einer Störung während der Kommunikation kann es jedoch vorkommen das ein Datensatz zweimal ausgelesen wird 
    (Aussage gilt für Firmwareversion > 2.0).
    Die Datensätze werden ab der Firmwareversion 03.01.20 beim AEIII+ sowie 04.00.21 beim PZE-MasterIV in der chronologischen Reihenfolge geliefert.

    \section HintForRecordAndListdefinitions Anmerkung zur Datensatz- und Listenbeschreibung (Tabellendefinitionen)
    Es können gleichzeitig mehrere Verbindungen, so genannte Kanäle, unter verwendung verschiedenen Schnittstellen aufgebaut werden. Für jeden Kanal wird ein eigener interner Speicherbereich verwendet, 
    der jedoch die Datensatzbeschreibungen und Listenbeschreibungen des zuletzt mit DFCLoadDatensatzbeschreibungausgelesenen Gerätes ( und DFCLoadListenbeschreibung) vorhält. D. h., wenn Sie verschiedene 
    Setupkonfigurationen in den Geräten verwenden, ist eine Aktualisierung der Datensatz- und Listenbeschreibungen erforderlich. 
    
    \section HintByOpenConnections Bei offen gehaltenen Kanälen ist folgendes zusätzlich zu beachten.
    Wenn die Verbindung zwischen der DLL und dem Gerät unterbrochen wurde, dann sollte dieses durch die Fehlerbehandlung auf folgende Weise abgefangen werden (siehe auch 1.2 Exemplarischer Programmaufbau). 
    * Prüfen ob alle anzusprechenden Geräte im Kanal einen Fehler bei "eigentlich" geöffnetem Kanal liefern. 
    * Die Verbindung des Kanals mit DFCComClose() schließen. 
    * Neuverbindung durch DFCComOpenIV() oder DFCComOpenSerial, DFCComOpenSocket je nach Gerätetyp und Verbindungsart durchführen.
*/

/*!
    \page CommonHints Allgemeine Hinweise

    Der C-Compiler dekoriert die in den Funktionsprototypen angegebenen Namen. Vor den Namen wird das Zeichen "_" gesetzt und an den Namen wird die Anzahl der übergebenen Parameter in Längen - Byte 
    durch "@" getrennt angehängt. Zusätzlich zum Funktionsprototyp werden daher auch die dekorierten Namen angegeben. Sie werden z.B. bei Verwendung von Visual Basic oder Delphi zum Import benötigt.

    Beim Arbeiten mit der DLL ist es unter Umständen hilfreich, das Tool "Depends" oder "Dependency Walker" einzusetzen. So können alle in der DLL vorhandenen Funktionen und die erwarteten Parameterbytes 
    angezeigt werden. Dieses Tool ist im Umfang von Visual Studio 6.0 enthalten. Dieser Export ist weiter unten bei jeder Funktion angegeben.

    Die Datei DFComDll.lib muss in den Projekteinstellungen bei den Optionen "Link" in Visual C++ Projekten bei den Objekt / Bibliothek-Modulen eingetragen sein und beim Kompilieren des Projektes im 
    Projektverzeichnis liegen (oder in dem in den Einstellungen eingestellten Pfad). Bei der Programmierung in C++ kann die Headerdatei aus dem Beispielcode verwendet werden. Auch für Visual Basic 
    ist ein Modul mit den notwendigen Exportangaben vorhanden.

    \section ListsOfSetup Listen des Setups

    \subsection ListsOfSetup1 Speicherverwaltung der Listen
    Jeder Kanal verwaltet seinen eigenen Speicher für die 20 möglichen Listen. Für jede zu importierende Liste wird entsprechend ihrer Größe ein dynamischer Speicher reserviert.
    Mit der Funktion DFCClrListenBuffer werden alle reservierten Speicher im entsprechenden Kanal freigegeben. 

    \attention Bitte beachten Sie, daß in den MasterIV-Geräten ein Flashspeicher zum Einsatz kommt. Laut dem Hersteller kann jeder Speicherblock (512Byte) maximal 100.000 mal beschrieben werden. 
    Die Firmware der Geräte verteilt die Zugriffslast auf die einzelnen Speicherblöcke und markiert intern defekte Blöcke. Es ist jedoch trotz dieses Sicherheitsmechanismus von all zu häufiger 
    Übertragung der Listendaten abzuraten. Die Applikation sollte nur bei Änderungen die Listen neu übertagen und nicht zyklisch permanent.  

    \subsection ListsOfSetup2 Übertragen von Listen
    Die Verarbeitung erfolgt in zwei Schritten. Zuerst wird der DLL interne Speicher durch einen Aufruf der Funktion DFCMakeListe für jede Liste mit Listendaten gefüllt. 
    Danach werden die Listendaten durch Aufruf der Funktion DFCLoadListen in das Gerät übertragen. Zu beachten ist, dass alle im Setup definierten Listen nacheinander mit der Funktion DFCMakeListe 
    in den Speicher der DLL übertragen werden. Durch Aufruf der Funktion DFCLoadListen werden dann alle zuvor importieren Listen auf einmal übertragen und alle entsprechend alten Listen im Gerät 
    automatisch gelöscht.

    \attention Zwischen der Übertragung von Listen muss ggf. eine Wartezeit eingehalten werden. Hierzu siehe Abschnitt: Beschreibung der Übernahme von Listen 

    \subsection ListsOfSetup3 Übertragen einzelner Listen
    Um eine einzelne Liste zu übertragen importieren Sie diese mit DFCMakeListe und übertragen sie dann mit DFCLoadListen. Die Funktion DFCLoadListen überträgt immer alle zuvor importierten Listen, egal ob 
    eine oder mehrere.

    \attention Das Laden einzelner Listen ist bei der MasterIV - Serie erst ab Firmwareständen >= 04.01.00 möglich. Beim AEIII+ müssen immer alle Listen übertragen werden, das Laden einzelner Listen ist 
    nicht möglich. 

    \attention Zwischen der Übertragung von Listen muss ggf. eine Wartezeit eingehalten werden. Hierzu siehe Abschnitt: Beschreibung der Übernahme von Listen 

    \subsection ListsOfSetup4 Löschen von Listen
    Es ist nicht möglich einzelne Listen komplett zu entfernen. Dies geschieht lediglich nach Übertragung einer Setupdatei. Durch Aufruf der Funktion DFCMakeListe mit der Datensatzanzahl 0 wird die Liste in 
    der DLL als Leere Liste importiert und durch Aufruf der Funktion DFCLoadListen wird diese dann übertragen.


    \subsection ListsOfSetup5 Beschreibung der Übernahme von Listen
    Beschreibung wie die Listenübernahme der Firmware arbeitet. Die Listenübernahme des Setups ist je nach gewähltem Betriebsmodus "Normal, MDE" oder "PZE Modus 1, 2" leicht verschieden.

    \subsubsection ListsOfSetup51 Listen für Betriebsmodus "Normal, MDE"
    Nach übertragen einer oder mehrerer Listen, werden diese in folgenden Situationen übernommen:

    Ein begonnener Menüeintrag (ob im Hauptmenü oder Untermenü) wird abgeschlossen, 
    Das Gerät befindet sich im Menü (ob im Hauptmenü oder Untermenü) selbst. 
    Listen für den Betriebsmodus "PZE Modus 1, 2"
    Zusätzlich zu den Situationen des Betriebsmodus "Normal, MDE", die hier auch gelten, kommt noch der einstellbare Tastentimeout dazu. Erfolgt keine Bedienung des Geräts für den eingestellten Tastentimeout, 
    durchläuft das Gerät zwangsläufig das Hauptmenü und die Listendaten werden dadurch übernommen.

    \subsubsection ListsOfSetup52 Listen für die Zutrittskontrolle
    Nach übertragen einer oder mehrerer Zutrittslisten wird die Zutrittskontrolle gestoppt und die Listendaten in das System übernommen. Danach nimmt die Zutrittskontrolle ihren Betrieb wieder auf.

    \subsubsection ListsOfSetup53 Listen für Timeboys
    Nach übertragen einer oder mehrerer Listen für Timeboys, werden diese sofort übernommen, sofern nicht gerade eine Übertragung der bereits vorliegenden Listen in einen Timeboy aktiv ist. Dann wird die 
    aktuelle Übertragung zuerst abgeschlossen bevor die neuen Listendaten übernommen werden.

    \subsection ListsOfSetup6 Beim AEIII+ sowie MasterIV Firmwareversionen < 04.01.00
    Hier müssen immer alle definierten Listen des Setups importiert und übertragen werden.

    Der DLL interne Puffer wird für jede "ComNum" getrennt geführt. Die Listendaten müssen daher auch für jede verwendete "ComNum" getrennt aufgebaut werden. 

    Wenn an einer "ComNum" mehrere Geräte angeschlossen sind, die unterschiedliche Listen benötigen, müssen die Listendaten vor der Übertragung zu der entsprechenden Busnummer, deren Gerät eine 
    unterschiedene Listenstruktur hat, neu in den Puffer der DLL geladen werden.

    Ablauf des Listenladen:

    * Alte Listen im Buffer mit "DFCClrListenBuffer" löschen. 
    * Neue Listen durch ein oder mehrfachen Aufruf von "DFCMakeListe" anlegen. 
    * Listen mit "DFCLoadListen" in das Gerät mit der entsprechenden Busnummer laden. 
    
    Testsoftware für die Listen liegt in folgenden Dateien bei: 

    * "DllTest.exe" Einfaches Testprogramm um eine Liste in ein Gerät zu laden. 
    * "DllTestSource.zip" Quelldateien für das Testprogramm (Visual C++ 6.0).  
    * "AepTSetup.aes" Setupdatei für das Testprogramm (muss vor dem Test in das Gerät geladen werden!). 
    
    Die Testliste besteht aus den 3 Namen: 
    Mueller, Maier und Schmidt mit den Personalnummern: 101, 102 und 103.


*/

/*!
    \page errorNumber Fehlernummer

    \section descriptionErrorNumber Beschreibung
    
    Die Fehlernummer ist ein wichtiger Parameter.
*/

/*!
    \page ChangeToConnectionIdAndDeviceAddress Änderung der Begriffe Kanalnummer zu Verbindungs-Id und Busnummer zu Geräteadresse ab Version 04.03.05.

    Änderung der Begriffe Kanalnummer zu Verbindungs-Id und Busnummer zu Geräteadresse ab Version 04.03.05. 
    Entwickler und alle anderen Interessierten, die bereits mit der DFCom - API vertraut sind, werden ab dieser Version ggf. die Begriffe Kanalnummer und Busnummer vermissen.
    In den beiden folgenden Kapiteln soll kurz auf die durchgeführte Umbenennung eingegangen werden.

    \section ChangeToConnectionId Änderung des Begriffs Kanalnummer zu Verbindungs-Id.

    Fast alle Funktionen der DFCom-API verwenden als ersten Parameter eine Verbindungs-Id. Diese gibt an, über welche Geräteverbindung die Funktion mit einem Gerät kommunizieren soll.
    Die Vergangenheit hat gezeigt, dass der Begriff Kanal für eine Verbindung immer wieder zu Missverständnissen geführt hat. Deswegen wurde entschieden, ab der API-Version 04.03.05 von einer Verbindung 
    zu sprechen. Um bei der API in sich konsistent zu sein, wurde durchgängig bei den Funktionen der Parametername channelID in connectionId umbenannt.

    \section ChangeToDeviceAddress Änderung des Begriffs Busnummer zu Geräteadresse.
    Auch die Geräteadresse wird bei fast allen Funktionen der DFCom-API als Parameter angegeben. Und zwar rührt sie aus Zeiten, als noch RS485 Busse die meisten Installationen beim Kunden bestimmten.
    Sie wird benötigt, um über eine Verbindung mit mehreren Geräten ein bestimmtes Gerät ansprechen zu können. Bei den Timeboy-Dockingstationen, die mehrere Einsteckplätze haben, kommt sie ebenfalls
    zum Einsatz, um anzugeben, auf welchen Einsteckplatz zugegriffen werden soll. Somit stellt die Geräteadresse zusammen mit der Verbindungs-Id einen eindeutigen Kommunikationsendpunkt dar.
    
    Auch bei der Busnummer hat die Vergangenheit gezeigt, dass immer wieder Missverständnisse aufgetreten sind. Auch hier wurde, um bei der API konsistent zu sein, durchgängig bei den Funktionen der 
    Parametername deviceID in deviceAddress umbenannt. 
*/

/*!
    \page ActiveMode Active-Mode

    \section ActiveModeIntroduction Einleitung
*/

/*!
    \page ImplemetationForOS Umsetzung des Kommunikationsmoduls für unterschiedliche Betriebssysteme.

    \section ImplementForWindows Umsetzung für Windows

    \section ImplementForUnix Umsetzung für unixartige Betriebssysteme.

    \subsection ImplementInStaticOrShared Erstellen einer statischen oder shared Library
*/

/*!
    \page DifferencesOfx86Andx64 Unterschiede zur Umsetzung des Kommunikationsmoduls für x64 Windows-Betriebssysteme.

*/

/*!
    \page ExportetFunctions Exportierte Funktionen

    \section CommonFunkctions Allgemeine Funktionen
*/

/*! 
    \brief Pruefen ob das Geraet erreichbar ist.

    Ruft intern die Funktion DFCCheckDevice() mit standard - Parametern auf.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.

    \return Ergebnis der Erreichbarkeit.
    \retval 0 Geraet ist nicht erreichbar.
    \retval 1 Es wurde erfolgreich auf das Geraet zugegriffen.
*/
DF_API_EXTERN int DF_API_CALL DFCCheckAE( int connectionId, int deviceAddress );



#if (DF_API_VERSION >= 0x02000300)


/*! 
    \brief Pruefen ob das Geraet erreichbar ist.

    Prueft aufgrund einer kleinen Datenanfrage die Erreichbarkeit des Geraetes.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.
    \param devicePollRetry Anzahl der Wiederholversuche.

    \return Ergebnis der Erreichbarkeit.
    \retval 0 Geraet ist nicht erreichbar.
    \retval 1 Es wurde erfolgreich auf das Geraet zugegriffen.
*/
DF_API_EXTERN int DF_API_CALL DFCCheckDevice( int connectionId, int deviceAddress, int *errorNumber, int devicePollRetry );

#endif


/*! 
    \brief Setzt das uebergebene Datum und die Uhrzeit in dem Geraet.

    Das zuvor aufbereitete Datum mit Uhrzeit muss in einem speziellen Format
    vorliegen, welches in der Dokumentation zur DLL beschrieben ist.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param dateTimeBuffer Zeiger auf Bytearray mit Werten fuer Datum und Uhrzeit.

    \return Ergebnis der Wertsetzung im Geraet.
    \retval 0 Wert konnte nicht gesetz werden.
    \retval 1 Wert wurde erfolgreich gesetzt.
*/
DF_API_EXTERN int DF_API_CALL DFCComSetTime( int connectionId, int deviceAddress, unsigned char *dateTimeBuffer );


/*! 
    \brief Holt das Datum und die Uhrzeit aus dem Geraet.

    Das Datum mit Uhrzeit wird in einem speziellen Format
    abgelegt, welches in der Dokumentation zur DLL beschrieben ist.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param dateTimeBuffer Zeiger auf Bytearray mit Werten fuer Datum und Uhrzeit.

    \return Ergebnis der Wertermittlung aus dem Geraet.
    \retval 0 Wert konnte nicht ermittelt werden.
    \retval 1 Wert wurde erfolgreich ermittelt.
*/
DF_API_EXTERN int DF_API_CALL DFCComGetTime( int connectionId, int deviceAddress, unsigned char *dateTimeBuffer );


/*! 
    \brief Direktes schreiben einer Textnachricht.

    Der gesendete Text wird fuer eine bestimmbare Zeit und mit einer
    eventuell akustischen Signalisierung auf der Anzeige ausgegeben. Die
    Ausgabe ist vom Geraetezustand abhaengig.
    
    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param visibleTime Anzeigedauer. (ucDelayTime    250ms). 255->nicht automatisch wegblenden.
    \param doAction Ist fuer eingriffe in die Bedienung vorgesehen.
    \param audibleSignal Meldung aktustisch untermalen.
    \param text Zeiger auf den anzuzeigenden Text.
    \param textLength Laenge des anzuzeigenden Textes.

    \return Ergebnis der Textsendung an Geraet.
    \retval 0 Text konnte nicht geschrieben werden.
    \retval 1 Text wurde erfolgreich geschrieben.
*/
DF_API_EXTERN int DF_API_CALL DFCComSendMessage( int connectionId, int deviceAddress, unsigned char visibleTime, unsigned char doAction, unsigned char audibleSignal, char *text, int textLength );


/*! 
    \brief Indirektes schreiben einer Textnachricht.

    Der gesendete Text wird in einem eigens dafuer vorgesehenen Speicherbereich 
    abgelegt und kann ueber die Pfeiltasten in den "nicht PZE-Moden" erreicht werden.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param text Zeiger auf den anzuzeigenden Text.
    \param textLength Laenge des anzuzeigenden Textes.

    \return Ergebnis der Textsendung an Geraet.
    \retval 0 Text konnte nicht geschrieben werden.
    \retval 1 Text wurde erfolgreich geschrieben.
*/
DF_API_EXTERN int DF_API_CALL DFCComSendInfotext( int connectionId, int deviceAddress, char *text, int textLength );


/*! 
    \brief Die Seriennummer aus dem Geraet lesen.

    Gibt die gelesene Seriennummer als Ganzzahl ueber Parameter zurueck.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.
    \param value Zeiger auf Variable welche die Seriennummer erhaelt.

    \return Ergebnis der Werteermittlung.
    \retval 0 Wert konnte nicht ermittelt werden und Fehlercode wurde ueber errorNumber abgelegt.
    \retval 1 Seriennummer wurde ueber piSerialNum in Variable uebergeben.
*/
DF_API_EXTERN int DF_API_CALL DFCGetSeriennummer( int connectionId, int deviceAddress, int *errorNumber, int *value );


#if (DF_API_VERSION >= 0x02000000)


/*! 
    \brief Setzen einer Rueckruffunktion.

    Die Rueckruffunktion wird waehrend des Kommunikationsablaufs fortwaehrend 
    aufgerufen und bietet die Moeglichkeit eines fruehzeitigen abbruches des selbigen.

    \param connectionId Kommunikationskanalnummer.
    \param functionPointer Zeiger auf die Rueckruffunktion oder 0 fuer abschalten.

    \return Keine.

    \version Ab DLL - Version 2.0.0 enthalten.
*/
DF_API_EXTERN void DF_API_CALL DFCSetCallBack( int connectionId, int (*functionPointer)(void) );

#endif


#if (DF_API_VERSION >= 0x02000300)
/*! 
    \brief Aufloesen einer Fehlernummer in seine lesbare Bedeutung.

    Die von den meisten Funktionen ueber den Parameter errorNumber zurueckgegebene
    Fehlernummer kann mit dieser Funktion in seine lesbare Bedeutung umgewandelt werden.

    \param connectionId Kommunikationskanalnummer.
    \param errorNumber Fehlernummer.
    \param languageId Angabe der zu verwendenden Sprache.
    \param value Zeiger auf die Zeichenfolge, welche den Fehlertext erhaelt.
    \param valueLength Laenge in Anzahl Byte des Speicherbereichs auf den szText verweist.

    \return Keine.

    \version Ab DLL - Version 2.0.3 enthalten.
*/
DF_API_EXTERN void DF_API_CALL DFCGetErrorText( int connectionId, int errorNumber, int languageId, char *value, int valueLength );

#endif

#if (DF_API_VERSION >= 0x02000400)


/*! 
    \brief Setzen des Wertes einer Globalen Variablen.

    Der Wert einer Globalen Variablen kann mit Hilfe dieser Funktion vom PC aus
    abgeaendert werden. Diese muss natuerlich im verwendeten Geraetesetup definiert sein.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param name Zeiger auf Byte welches die logische Variablennummer enthaelt, oder
                  auf eine Zeichenfolge, welche den Variablennamen darstellt.
    \param nameType Gibt an ob name auf eine Variablennummer oder auf einen Variablennamen verweist.
    \param value Zeiger auf den Wert der gesetzt werden soll.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Wertuebernahme im Geraet.
    \retval 0 Wert konnte nicht uebernommen werden.
    \retval 1 Wert wurde erfolgreich uebernommen.

    \version Ab DLL - Version 2.0.4 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCSetGlobVar( int connectionId, int deviceAddress, unsigned char *name, int nameType, unsigned char *value, int *errorNumber );


/*! 
    \brief Lesen des Wertes einer Globalen Variablen.

    Der Wert einer Globalen Variablen kann mit Hilfe dieser Funktion vom PC aus
    gelesen werden. Diese muss natuerlich im verwendeten Geraetesetup definiert sein.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param name Zeiger auf Byte welches die logische Variablennummer enthaelt, oder
                  auf eine Zeichenfolge, welche den Variablennamen darstellt.
    \param nameType Gibt an ob name auf eine Variablennummer oder auf einen Variablennamen verweist.
    \param value Zeiger auf ermittelte nullterminierte Zeichenfolge.
    \param valueLength Laenge in Anzahl Byte des Speicherbereichs auf den pucValue verweist.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Wertermittlung aus Geraet.
    \retval 0 Wert konnte nicht ermittelt werden.
    \retval 1 Wert wurde erfolgreich ermittelt.

    \version Ab DLL - Version 2.0.4 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCGetGlobVar( int connectionId, int deviceAddress, unsigned char *name, int nameType, unsigned char *value, int valueLength, int *errorNumber );

#endif

#if (DF_API_VERSION >= 0x02000500)


/*! 
    \brief Schliessen eines angeschlossenen Relais.

    Internen digitalen Ausgang oder externen digitalen Ausgang (z.B. IOModule)
    fuer eine bestimmbare Zeit schalten.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param number Gibt die Nummer des zu schaltenden Relais an.
    \param timeout Gibt die Zeit in Sekunden an, in welcher das Relais geschaltet bleibt.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden.
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 2.0.5 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCCloseRelay( int connectionId, int deviceAddress, int number, int timeout, int *errorNumber );


/*! 
    \brief Abfragen von aktuellem Zustand und verbleibender Zeit.

    Der aktuelle Zustand und die verbleibende Zeit bei geschlossenem Relais abfragen.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param number Gibt die Nummer des zu schaltenden Relais an.
    \param state Zeiger auf eine Variable, welche den aktuellen Schaltzustand des Ausgangs erhaelt.
    \param timeout Zeiger auf die Variable, welche die verbleibende Zeit in Sekunden erhaelt.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden.
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 2.0.5 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCGetRelayState( int connectionId, int deviceAddress, int number, int *state, int *timeout, int *errorNumber );


/*! 
    \brief Oeffnen eines Relais.

    Internen digitalen Ausgang oder externen digitalen Ausgang (z.B. IOModule) schalten.

    \param connectionId Kommunikationskanalnummer.
    \param deviceAddress Busnummer des anzusprechenden Geraetes.
    \param number Gibt die Nummer des zu schaltenden Relais an.
    \param errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.

    \return Ergebnis der Ausfuehrung.
    \retval 0 Funktion konnte nicht ausgefuehrt werden.
    \retval 1 Funktion wurde erfolgreich ausgefuehrt.

    \version Ab DLL - Version 2.0.5 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCOpenRelay( int connectionId, int deviceAddress, int number, int *errorNumber );
#endif

#if (DF_API_VERSION >= 0x02000500)
/*! 
    \brief Abfrage der zuletzt mit DFCCheckDevice() benoetigten Versuche.

    Die Anzahl der benoetigten Versuche koennen fuer eventuelle Statistikzwecke 
    verwendet werden.

    \param connectionId Kommunikationskanalnummer.

    \return Anzahl der benoetigten Versuche bis Geraet als verfuegbar galt.

    \version Ab DLL - Version 2.0.5 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCGetDevicePollRetry( int connectionId );

#endif


#if (DF_API_VERSION >= 0x02001800)

DF_API_EXTERN int DF_API_CALL DFCGetVersionFirmware( int connectionId, int deviceAddress, char *value, int *errorNumber );
#endif

#if (DF_API_VERSION >= 0x02002100)

DF_API_EXTERN int DF_API_CALL DFCGetVersionFirmwareFromFile( int connectionId, char *filePath, char *value, int *errorNumber );
#endif

#if (DF_API_VERSION >= 0x04020300)

/*! 
* \brief Die Hardware Info vom Geraet lesen.
*
* Liest die Hardware Info aus dem Geraet und schreibt sie in eine anzugebende Ini-Datei.
*
* \param[in] connectionId Kommunikationskanalnummer.
* \param[in] deviceAddress Busnummer des anzusprechenden Geraetes.
* \param[out] value Zeiger auf den String fuer die Hardwareinfo.
* \param[in] valueLength Zeiger auf die Laengenangabe des Strings.
* \param[in] flags Zusatzangaben. (noch unverwendet)
* \param[out] errorNumber Zeiger auf Variable die nach Fehler den Fehlercode erhaelt.
*
* \return Ergebnis der Werteermittlung.
* \retval 1 Werte wurden ermittelt und in Ini-Datei abgelegt.
* \retval 0 Fehler bei Werteermittlung, Fehlernummer ueber Parameter errorNumber ermittelbar.
*/
DF_API_EXTERN int DF_API_CALL DFCReadHardwareInfo( int connectionId, int deviceAddress, char *value, int *valueLength, int flags, int *errorNumber );
#endif

#if (DF_API_VERSION >= 0x04010600)
DF_API_EXTERN int DF_API_CALL DFCSetOptionFirmware( int connectionId, int deviceAddress, int mask, int options, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCGetOptionFirmware( int connectionId, int deviceAddress, int *mask, int *options, int *errorNumber );
#endif

#if (DF_API_VERSION >= 0x04010700)
DF_API_EXTERN int DF_API_CALL DFCSetFontType( int connectionId, int deviceAddress, int type, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCPressVirtualKey( int connectionId, int deviceAddress, int keyCode, int flags, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCGetPasswordKey( int connectionId, int deviceAddress, char *value, int *valueLength, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCSetPassword( int connectionId, char *value, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCReset( int connectionId, int deviceAddress, int mode, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCGetFlashStatus( int connectionId, int deviceAddress, int *value, int *errorNumber );

#endif

#if (DF_API_VERSION >= 0x04020300)
DF_API_EXTERN int DF_API_CALL DFCSetCommunicationPassword( int connectionId, unsigned char *value, int valueLength, int sessionTimeout, int *errorNumber );
#endif

DF_API_EXTERN int DF_API_CALL DFCRecordVolume( int connectionId, int deviceAddress, int *recordCount, int *memoryUsage, int *memorySize, int *errorNumber );


#if (DF_API_VERSION >= 0x04010400)
DF_API_EXTERN int DF_API_CALL DFCGetInfo( int connectionId, int deviceAddress, char *name, char *params, char *text, int *textLength, int *errorNumber );

DF_API_EXTERN int DF_API_CALL DFCOpenComServerMode( int connectionId, int deviceAddress, char *comm, int baudRate, int byteSize, int parity, int stopBits, int flags, int timeoutToClose, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCCloseComServerMode( int connectionId, int deviceAddress, int *errorNumber );
#endif

// ########################################################
// ########### FUNKTIONEN FUER ERWEITERUNGEN ##############
// ########################################################

#if (DF_API_VERSION >= 0x04030001)


/*!
* \brief Ermitteln der Libraryversion.
*
* Die Version wird im Format XX.XX.XX.XX geliefert.
*
* \param value Zeiger auf Zeichenfolgenpuffer für die Version.
* \param valueLength Länge des reservierten Puffers für die Zeichenfolge. Muss mindestens 12 sein.
*
* \return Ergebnis der Ausfuehrung.
* \retval 0 Parameter außerhalb des Gültigkeitsbereichs.
* \retval 1 Funktion wurde erfolgreich durchgefuehrt.
*
* \version Ab DLL - Version 04.03.00 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCVersionLibrary( char *value, int valueLength );

#endif

#if (DF_API_VERSION >= 0x04030500)


/*!
* \brief Ermitteln der Libraryversion.
*
* Die Version wird im Format XX.XX.XX.XX geliefert.
*
* \param value Zeiger auf Zeichenfolgenpuffer für die Version.
* \param valueLength Länge des reservierten Puffers für die Zeichenfolge. Muss mindestens 12 sein.
*
* \return Ergebnis der Ausfuehrung.
* \retval 0 Parameter außerhalb des Gültigkeitsbereichs.
* \retval 1 Funktion wurde erfolgreich durchgefuehrt.
*
* \version Ab DLL - Version 04.03.05 enthalten.
*/
DF_API_EXTERN int DF_API_CALL DFCTransparencyMode( int connectionId, int deviceAddress, int id, int flags, unsigned int sizeBufferRead, unsigned int sizeBufferWrite, int timeout, int *errorNumber );
DF_API_EXTERN int DF_API_CALL DFCIsTransparencyMode( int connectionId );
#endif


#if (DF_API_VERSION >= 0x04030600)

/*!
* \brief Ermitteln der Libraryversion.
*
* Die Version wird im Format XX.XX.XX.XX geliefert.
*
* \param value Zeiger auf Zeichenfolgenpuffer für die Version.
* \param valueLength Länge des reservierten Puffers für die Zeichenfolge. Muss mindestens 12 sein.
*
* \return Ergebnis der Ausfuehrung.
* \retval 0 Parameter außerhalb des Gültigkeitsbereichs.
* \retval 1 Funktion wurde erfolgreich durchgefuehrt.
*
* \version Ab DLL - Version 04.03.06 enthalten.
*/

DF_API_EXTERN int DF_API_CALL DFCCustomCommand( int connectionId, int deviceAddress, int identification, int customCommand, char *buffer, int *bufferSize, int *errorNumber );

#endif


// ########################################################
// #############* PRIVATE FUNKTIONEN* #################
// ########################################################

// Funktionen sind nicht dokumentiert.
// Sie werden von AESetup oder Talk verwendet.
// Unterliegen evntl. Aenderungen.

#if defined( USE_PRIVATE )

#if defined( _MSC_VER ) && defined( DF_API_EXPORT ) && (DF_API_VERSION >= 0x02000000)

HRESULT DllGetVersion( DLLVERSIONINFO *pdvi );

#endif

#if (DF_API_VERSION >= 0x02000200)

DF_API_EXTERN int DF_API_CALL DFCComVersion( int connectionId, int deviceAddress, int *errorNumber );

#endif

#if (DF_API_VERSION >= 0x03020000)

DF_API_EXTERN void* DF_API_CALL DFCGetClass( int connectionId, int classId );

#endif

#endif // USE_PRIVATE

#endif


