/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file is part of the DFComDLL module of the Datafox SDK.
**
****************************************************************************/

#include "Channel4Protocol.h"
#include <stdio.h>
#include <stdlib.h>
#include <boost/format.hpp>
#include <memory.h>

// ##################################################################
/**
* @brief Initialisieren der Membervariablen.
*
* @param channelId Nummer der Kanalklasse fuer z.B. Trace.
* @param deviceId Busnummer der Kanalklasse.
*/
Channel4Protocol::Channel4Protocol( short channelId, short deviceId ) : Channel4Base( channelId )
{
    // Umgebung zuruecksetzen
    reset();

    // DeviceID der Klasse zuweisen
    _deviceId = deviceId;
    _deviceIdRemote = IV_DEFAULT_DEVICEID_STANDALONE;

    // Zustand des Kanalobjekts
    _activeState = is_free;

    // Puffer zur schnellen Uebertragung
    memset( _tmp, 0, sizeof( _tmp ) );
    _tmpPos = _tmp;

    // Die Umgebung des verwendeten CRC32 initialisieren
    initCrc32Table();

    // Membervariablen fuer Paketauswertung initialisieren
    _packetStart = FALSE;
    _skipCount = 0;
    _packetNumber = 1;
    _lastFlags = 0;

    _clocksWrite = _clockWrite = 0;
    _clocksRead = _clockRead = 0;

    _packetWriteCounter = 0;
    _writtenSize = 0;
    _packetWritePayloadSize = 0;

    _packetReadCounter = 0;
    _packetLostCounter = 0;
    _amountSize = 0;
    _packetReadPayloadSize = 0;

    _qs = 0;

    // Callbacks intialisieren
    _callBackFunction = NULL;
    _callBackFunctionWithParams = NULL;

    // Puffer fuer Senden, Empfang
    memset( _packetWrite, 0, sizeof( _packetWrite ) );
    memset( _packetRead, 0, sizeof( _packetRead ) );

    // Hilfsmember initialisieren
    _previousModule   = 0;
    _previousFunction = 0;
    _result = 0;

    // Zufallsgenerator starten
    srand( static_cast<unsigned int>( time( NULL ) ) );

    // Verschluesselung aktivieren
    df4Init( TRUE );

    // Kein Autobaud bei Verbindungsherstellung.
    setAutobaud( false );

    // Vorgegebene Version der Dockingstation. _timeboyStation wird nur bei Version == 1 verwendet.
    _timeboyStationEnabled = 1;

    // Kommunikationsbasis an TimeboyStation weitergeben
    _timeboyStation.setChannelBase( getChannelBase() );
}

// ##################################################################
/**
* @brief Destruktor.
*
*/
Channel4Protocol::~Channel4Protocol() 
{
}


//#########################################################
//####################### HELPER ##########################
//#########################################################

// ##################################################################
/**
* @brief Initialisiert die grundlegend benoetigten Membervariablen.
*/
void Channel4Protocol::reset()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    _result = IV_ERROR_SUCCESS;
    resetError();
    memset( _packetData, 0, sizeof( _packetData ) );
}

// ##################################################################
/**
* @brief Setzen einer auszufuehrenden remoteCall-Back Funktion.
*/
void Channel4Protocol::setCallBackFunction( int (*pFunction)() )
{	
    trace( "%s(%i):INFO(3): %s( %p )\n", __FILE__, __LINE__, __func__, pFunction );

    _callBackFunction = pFunction;
}

// ##################################################################
/**
* @brief Setzen einer auszufuehrenden remoteCall-Back Funktion mit Parametern.
*/
void Channel4Protocol::setCallBackFunctionWithParams( int (*pFunction)( unsigned int, unsigned int ) )
{
    trace( "%s(%i):INFO(3): %s( %p )\n", __FILE__, __LINE__, __func__, pFunction );

    _callBackFunctionWithParams = pFunction;
}

//#########################################################
//######################### CRC ###########################
//#########################################################

// ##################################################################
/**
* @brief Reflection is a requirement for the official CRC-32 standard.
* You can create CRCs without it, but they won't conform to the standard.
*/
unsigned int Channel4Protocol::reflect( unsigned int ref, char ch )
{
    // Used only by initCrc32Table()
    trace( "%s(%i):INFO(3): %s( %i, %i )\n", __FILE__, __LINE__, __func__, ref, ch );

    unsigned int val = 0;

    // Swap bit 0 for bit 7
    // bit 1 for bit 6, etc.
    int i;
    for( i = 1; i < (ch + 1); ++i )
    {
        if ( ref & 1 )
            val |= 1 << (ch - i);

        ref >>= 1;
    }

    return val;
}

// ##################################################################
/**
* @brief remoteCall this func only once to initialize the CRC table.
*/
void Channel4Protocol::initCrc32Table() 
{
    // Called by Init
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // This is the official polynomial used by CRC-32 
    // in PKZip, WinZip and Ethernet. 
    unsigned int ulPolynomial = 0x04c11db7;

    // 256 values representing ASCII character codes.
    int i, j;
    for( i = 0; i <= 0xFF; ++i )
    {
        _crc32_table[i] = reflect( i, 8 ) << 24;
        for ( j = 0; j < 8; ++j )
            _crc32_table[i] = (_crc32_table[i] << 1) ^ (_crc32_table[i] & (1 << 31) ? ulPolynomial : 0);

        _crc32_table[i] = reflect( _crc32_table[i], 32 );
    }
}

// ##################################################################
/**
* @brief This func uses the crc32_table lookup table to generate a CRC for data.
*/
unsigned int Channel4Protocol::crc32( unsigned char *data, unsigned int size, unsigned int crc, unsigned char closure )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, 0x%08X, %i )\n", __FILE__, __LINE__, __func__, data, size, crc, closure );

    // Be sure to use unsigned variables,
    // because negative values introduce high bits
    // where zero bits are required.
    //unsigned int crc = 0xffffffff;
    int len;
    unsigned char *buffer;

    len = size;
    // Save the text in the buffer.
    buffer = data;
    // Perform the algorithm on each character
    // in the string, using the lookup table values.
    while ( len-- )
        crc = (crc >> 8) ^ _crc32_table[(crc & 0xFF) ^ *buffer++];

    // Exclusive OR the result with the beginning val.
    return (closure) ? crc ^ 0xffffffff : crc;
}

// ##################################################################
/**
* @brief Erzeugt einen 16bit CRC Wert fuer die uebergebenen Daten.
*/
unsigned short Channel4Protocol::crc16( unsigned char *data, unsigned int size, unsigned short crc )
{
    trace( "%s(%i):INFO(3): %s( %p, %i, 0x%04X ]\n", __FILE__, __LINE__, __func__, data, size, crc );

    unsigned int i;
    unsigned char j;

    for ( i = 0; i < size; ++i )
    {
        crc = crc ^ data[i];

        for ( j = 0; j < 8; ++j )
        {
            if ( crc & 0x01 )
            {
                crc >>= 1; 
                crc = crc ^ 0xA001;
            }
            else 
            {
                crc >>= 1;
            }
        }
    }

    return crc;
}


//#########################################################
//################ Verschluesselung #######################
//#########################################################

void Channel4Protocol::df4LoadCryptKey()
{
    unsigned char cryptKey_DF4[32] = 
    {	
        0xc5, 0xc2, 0xc3, 0xc0, 0xc1, 0xd9, 0xdb, 0xf0, 
        0xcd, 0x07, 0xa7, 0xaf, 0xaa, 0x0a, 0x50, 0xae,
        0xc5, 0x4e, 0xc3, 0xd9, 0xaf, 0xd9, 0x09, 0xc1,
        0xa7, 0xef, 0xac, 0xc2, 0x0a, 0xbf, 0xcd, 0xae
    };

    memcpy( _key, cryptKey_DF4, sizeof(cryptKey_DF4) );
}

// ##################################################################
/**
* @brief Initialisiert die sbox fuer die DF4-Verschlusselungsroutine.
*/
void Channel4Protocol::df4Init( int binaryKey )
{
    unsigned int i;
    unsigned int j;
    unsigned char *kptr = _key, trans;

    for( i = 0; i < 256; ++i )
    {
        _sbox[i] = static_cast<unsigned char>( i );

        _k[i] = *kptr;
        kptr++;

        if ( binaryKey )
        {
            if( kptr == (_key + sizeof( _key )) )
                kptr = _key;
        }
        else
        {
            if ( *kptr == 0 )
                kptr = _key;
        }
    }

    for( i = 0, j = 0; i < 256; ++i )
    {
        j = (j + _sbox[i] + _k[i]) % 256;
        trans = _sbox[i];
        _sbox[i] = _sbox[j];
        _sbox[j] = trans;
    }
}

// ##################################################################
/**
* @brief Ver-/Entschluesselt die uebergebenen Daten.
*/
void Channel4Protocol::df4( unsigned char *data, int size )
{
    unsigned int i;
    unsigned int j;
    unsigned int t;
    unsigned int pos;
    unsigned char trans;

    // Kopie von sbox, da sie unten manipuliert wuerde.
    memcpy( _k, _sbox, sizeof( _k ) );

    // Wenn die Angabe einer Laenge fehlt, wird davon ausgegangen,
    // dass mit einer Zeichenkette gearbeitet wird.
    if ( size == 0 )
        size = static_cast<int>( strlen( reinterpret_cast<char *>( data ) ) );

    i = j = pos = 0;
    for( ; (size - pos) > 0; ++pos )
    {
        i = (i + 1) % 256;
        j = (j + _k[i]) % 256;
        trans = _k[i];
        _k[i] = _k[j];
        _k[j] = trans;
        t = (_k[i] + _k[j]) % 256;
        data[pos] ^= _k[t];
    }
}

// ##################################################################
/**
* @brief Prueft die Erreichbarkeit des Kommunikationspartners.
*
* @param deviceId Anzusprechender Kommunikationspartner.
*/
int Channel4Protocol::ping( short deviceId, unsigned char *data, int *count )
{
    trace( "%s(%i):INFO(3): %s( %i, %p, %p )\n", __FILE__, __LINE__, __func__, deviceId, data, count );
    int lCount = 0;
    int autobaudrate[5] = { 38400, 9600, 19200, 57600, 115200 };
    int loop;

    // Funktion in Remoteklasse aufrufen?
    if ( getDeviceId() != deviceId )
    {
        // Umgebung zuruecksetzen
        resetError();

        // Remoteadresse setzen.
        setDeviceIdRemote( deviceId );

        // Zu sendende Daten kopieren?
        if ( data && count && *count > 0 )
        {
            lCount = *count;
            memcpy( _packetData, data, lCount );
        }

        Channel4Base::enumType type;
        std::string commString;
        int commVal, oldCommVal, commTO, oldCommTO;
        getChannelBase().getOpenParams( type, commString, commVal, commTO );
        oldCommVal = commVal;
        oldCommTO = commTO;

        for ( loop = 0; loop < 3; ++loop )
        {
            // Bei Autobaud und Seriell wird die Kommunikation in folgender Reihenfolge abgefragt.
            // 1. Voreingestellter Baudwert.
            // Je nach Voreinstellung folgende Prio.
            // - 38400 bd
            // - 9600 bd
            // - 19200 bd
			// -57600 bd
			// -115200 bd
            if ( getAutobaud() && type == Channel4Base::Serial )
            {
                // Ab zweitem loop auf naechste Baudrate aus dem Pool wechseln.
                if ( loop > 0 )
                {
                    if ( autobaudrate[0] )
                        commVal = autobaudrate[0];
                    else if ( autobaudrate[1] )
                        commVal = autobaudrate[1];
                    else if ( autobaudrate[2] )
                        commVal = autobaudrate[2];
					else if ( autobaudrate[3] )
						commVal = autobaudrate[3];
					else if ( autobaudrate[4] )
						commVal = autobaudrate[4];
					else
                        break; // Abbruch der Schleife, wenn alle zu pruefenden durch sind.

                    // Schnittstellenbaud anpassen
                    getChannelBase().changeStateSerial( getChannelBase()._comm, commVal, commTO );
                }

                // Aktuell gewaehlte Baudrate aus dem Pool nehmen.
                if ( commVal == autobaudrate[0] )
                    autobaudrate[0] = 0;
                else if ( commVal == autobaudrate[1] )
                    autobaudrate[1] = 0;
                else if ( commVal == autobaudrate[2] )
                    autobaudrate[2] = 0;
				else if ( commVal == autobaudrate[3] )
					autobaudrate[3] = 0;
				else if ( commVal == autobaudrate[4] )
					autobaudrate[4] = 0;
			}
            else
            {
                // Kein Autobaud gewuenscht oder kein RS232.
                loop = 3;
            }

            if ( (_result = call( IV_CONTEXT_MASTER, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, _packetData, lCount, 2 )) == IV_ERROR_SUCCESS 
                || (_result == -IV_ERROR_NEED_USER_PASSWORD) )
                break;
        }

        // Resultwert hier abpruefen!
        if ( _result != IV_ERROR_SUCCESS )
        {
            // Nach Fehlgeschlagener Ermittlung, urspruengliche Einstellung wiederherstellen.
            if ( getAutobaud() && type == Channel4Base::Serial )
                getChannelBase().changeStateSerial( getChannelBase()._comm, oldCommVal, oldCommTO );

            // Fehler ausgeben
            traceError( deviceId, IV_MODULE_PROTOCOL, IV_FUNCTION_PING );
            return _result;
        }

        // Empfangene Daten kopieren?
        if ( count )
        {
            *count = lCount;
            if ( data && *count > 0 )
            {
                lCount = *count;
                memcpy( _packetData, data, lCount );
            }
        }
    }
    else
    {
        int result;

        // Parameterdaten werden beibehalten

        // Antwort des Befehls senden.
        if ( (result = call( IV_CONTEXT_TRANSMIT_ANSWER, IV_MODULE_PROTOCOL, IV_FUNCTION_PING, data, lCount )) != IV_ERROR_SUCCESS )
        {			
            trace( "%s(%i):ERROR(1): %s() [ call (%i) ]\n", __FILE__, __LINE__, __func__, result );
            return result;
        }
    }

    return _result;
}

// ##################################################################
/**
* @brief Schreiben eines teilaufbereiteten Packets in die Kommunikationsschnittstelle.
* Der Kopf und Packet-CRC wird vor Packetversand gebildet und zur Packetgroesse dazugerechnet
* und nach Versand wieder entfernt.
*/
int Channel4Protocol::writeProtocol( unsigned char *buffer, int writeTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %i )\n", __FILE__, __LINE__, __func__, buffer, writeTrials );

    int res = IV_ERROR_SUCCESS;
    int transmittedCount = 0, transmitted, enableMasking = 1, masked = 0;

    // Laengenangabe pruefen
    if ( *reinterpret_cast<unsigned short *>( &buffer[4] ) < IV_PACKET_LENGTH_FRAME
        && *reinterpret_cast<unsigned short *>( &buffer[4] ) >= IV_PACKET_LENGTH_MAX )
    {
        trace( "%s(%i):ERROR(1): %s() [ packet length fail: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_PACKETLENGTH );
        return -IV_ERROR_PACKETLENGTH;
    }

    // Temppuffer setzen
    _tmpPos = _tmp;

    // Handelt es sich um eine Dockingstation?
    if ( isTimeboySlot( getDeviceIdRemote() ) )
    {
        // Keine Maskierung der Paketdaten bei versenden ueber eine Dockingstation.
        enableMasking = 0;
    }

    // CRC - groesse daraufrechnen
    *reinterpret_cast<unsigned short *>( &buffer[4] ) += sizeof( unsigned short );
    // CRC von Paketkopf einfuegen
    *reinterpret_cast<unsigned short *>( &buffer[8] ) = crc16( buffer, IV_PACKET_LENGTH_HEAD - sizeof( unsigned short ) );
    // CRC von Paket anfuegen
    *reinterpret_cast<unsigned short *>( &buffer[*reinterpret_cast<unsigned short *>( &buffer[4] ) - sizeof( unsigned short )] ) 
        = crc16(buffer, *reinterpret_cast<unsigned short *>( &buffer[4] ) - sizeof( unsigned short ) );
    /*
    if (IV_PACKET_MODULE(buffer) != 0) 
    {
    // CRC von Paket fälschen
    if (!(rand() % 8)) 
    {
    *(unsigned short*)&buffer[*(unsigned short*)&buffer[4] - sizeof(unsigned short)] += 1;
    Trace("%s(%i):ERROR(1): %s [ verfälsche gesamt CRC ]\n", __FILE__, __LINE__, __func__);
    }
    }
    */
    
    // Ggf. Packetdaten raustracen.
    if ( isTrace() >= 2 )
    {
        trace( "%s(%i):INFO(1): %s() ", __FILE__, __LINE__, __func__ );
        tracePacket( buffer );
    }


    // Datenstrom zeichenweise bearbeiten und besondere Zeichen maskieren
    for ( ; transmittedCount < *reinterpret_cast<unsigned short *>( &buffer[4] ); ++transmittedCount )
    {	
        // Paketzeichen im Datenstrom mit Escapesequenz umschreiben
        if ( enableMasking && (transmittedCount > 1) 
            && ((buffer[transmittedCount] == IV_PACKET_SIGN_START)
            || (buffer[transmittedCount] == IV_PACKET_SIGN_ESCAPE)) )
        {
            unsigned char replace[2];
            replace[0] = IV_PACKET_SIGN_ESCAPE;
            // Anzahl maskierte mitzaehlen
            masked++;

            switch ( buffer[transmittedCount] ) 
            {
            case IV_PACKET_SIGN_START:
                replace[1] = 0x01;
                break;

            case IV_PACKET_SIGN_ESCAPE:
                replace[1] = 0x00;
                break;
            }

            // Wuerde Pufferende erreicht?
            if ( ((_tmpPos + sizeof( replace )) - _tmp) > static_cast<int>( sizeof( _tmp ) ) ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ Puffer overflow: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_TRANSMIT );
                return -IV_ERROR_TRANSMIT;
            }

            *_tmpPos = reinterpret_cast<unsigned char *>( replace )[0];
            _tmpPos++; // Auf naechstes Zeichen

            *_tmpPos = reinterpret_cast<unsigned char *>( replace )[1];
            _tmpPos++; // Auf naechstes Zeichen

        } 
        else 
        {
            // Wuerde Pufferende erreicht?
            if ( ((_tmpPos + 1) - _tmp) > static_cast<int>( sizeof( _tmp ) ) ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ Puffer overflow: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_TRANSMIT );
                return -IV_ERROR_TRANSMIT;
            }

            *_tmpPos = buffer[transmittedCount];
            _tmpPos++; // Auf naechstes Zeichen
        }
    }

    // Absenden des Gesamtpaketes
    if ( transmittedCount == *reinterpret_cast<unsigned short *>( &buffer[4] ) ) 
    {
        _clockWrite = 0;
        // Anfang des Versendens halten
        clock_t clockStart = clock();
        if ( isTimeboySlot( getDeviceIdRemote() ) )
        {
            if ( (res = _timeboyStation.writeBytes( getDeviceIdRemote(), _timeboyBaseAddress, 0, _tmp, *reinterpret_cast<unsigned short *>( &buffer[4] ) + masked )) != IV_ERROR_SUCCESS )
            {
                trace( "%s(%i):ERROR(1): %s() [ writeBytes fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
                goto label_error;
            }
        }
        else
        {
            if ( (res = writeData( _tmp, *reinterpret_cast<unsigned short *>( &buffer[4] ) + masked, transmitted, writeTrials )) != IV_ERROR_SUCCESS ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ TransmitData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
                goto label_error;
            }
        }

        // Gesamtzeit des Versendevorgangs fuer Statistik halten
        _clockWrite = clock() - clockStart;
        _clocksWrite += _clockWrite;
    }

label_error:
    // CRC - groesse wieder abziehen
    *reinterpret_cast<unsigned short *>( &buffer[4] ) -= sizeof(unsigned short);
    // CRC von Paketkopf loeschen
    *reinterpret_cast<unsigned short *>( &buffer[8] ) = 0;
    // CRC von Paket loeschen
    *reinterpret_cast<unsigned short *>( &buffer[*reinterpret_cast<unsigned short *>( &buffer[4] )] ) = 0;

    return res;
}

// ##################################################################
/**
* @brief Schreiben eines Packets fuer die angegebene Modulfunktion.
*
* @param deviceIdRemote Anzusprechende Geraetenummer.
*/
// Aufbereiten und senden eines Anfrage-Pakets
//  deviceIdRemote -> Adresse (Geraetenummer) des Empfaengers
//  type     -> Pakettyp (Anfrage/Antwort +? Error)
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
int Channel4Protocol::writePacket( short deviceIdRemote, unsigned char type, short module, short func, unsigned char *buf, int count, int writeTrials )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i, %p, %i )\n", __FILE__, __LINE__, __func__, deviceIdRemote, module, func, buf, count );

    int err = IV_ERROR_SUCCESS;

    // CallBack-Function vor Packetversant aufrufen.
    if ( _callBackFunction )
    {
        if ( !_callBackFunction() ) 
        {
            trace( "%s(%i):ERROR(1): %s() [ CallBack abbruch ]\n", __FILE__, __LINE__, __func__ );
            return -IV_ERROR_BREAK;
        }
    }

    // Remoteadresse setzen
    setDeviceIdRemote( deviceIdRemote );

    /* 
    1.	(  8 Bit): Startzeichen ("~", 0x7E)
    2.	(  8 Bit): Empfängeradresse
    3.	(  8 Bit): Absenderadresse
    4.	(  8 Bit): Steuerinformationslänge
    5.	(16 Bit): Paketgesamtlänge
    6.	(  8 Bit): Flags (z. B. für aktive Verschlüsselung)
    7.	(  8 Bit): Reserve
    8.	(16 Bit): CRC über die ersten 7 Felder
    9.	(n*8 Bit): Steuerinformationsdaten (in Bezug auf Feld 5)
    10.	(n*8 Bit): Zu transportierende Daten
    11.	(16 Bit): CRC über Felder 1 - 10.
    */
    // 1.
    _packetWrite[0] = IV_PACKET_SIGN_START;
    // 2.
    _packetWrite[1] = static_cast<unsigned char>( getDeviceIdRemote() );
    // 3.
    _packetWrite[2] = static_cast<unsigned char>( getDeviceId() );
    // 4.
    _packetWrite[3] = IV_PACKET_LENGTH_CONTROLBLOCK;
    // 5.
    *reinterpret_cast<unsigned short *>( &_packetWrite[4] ) 
        = IV_PACKET_LENGTH_HEAD + _packetWrite[3] 
            + static_cast<unsigned short>( sizeof( unsigned char ) ) + static_cast<unsigned short>( sizeof( unsigned short ) ) * 2 + count;
    // 6.
    _packetWrite[6] = 0;
    // 7.
    _packetWrite[7] = 0;
    // 8.
    // Wird von Transmit - Funktion erledigt
    // 9.
    // Zu Punkt 9: Steuerinformation -> Packezaehler
    // Ist Packet eine Anfrage?
    if ( type & IV_PACKET_REQUEST ) 
    {
        // Aktuelle Packetnummer
        _packetWrite[IV_PACKET_LENGTH_HEAD] = _packetNumber;
    } 
    else if ( type & IV_PACKET_ANSWER ) 
    {
        // Packet ist Antwort auf Anfrage!
        // Steuerinformation verfuegbar? 
        if ( _packetRead[3] > 0 ) 
        {	
            // Dann in die Antwort uebernehmen.
            _packetWrite[IV_PACKET_LENGTH_HEAD] = _packetRead[IV_PACKET_LENGTH_HEAD];
        }
    }

    // 10.
    _packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3]] = IV_PACKET_VERSION_4 | type;
    *reinterpret_cast<unsigned short *>( &_packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3] + sizeof( unsigned char )] ) = module;
    *reinterpret_cast<unsigned short *>( &_packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3] + sizeof( unsigned char ) + sizeof( unsigned short )] ) = func;

    // Daten in Sendebuffer kopieren
    memcpy( &_packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3] + sizeof( unsigned char ) + sizeof( unsigned short ) * 2], buf, count );	

    // 10a
    // Daten Verschluesselung wird durch CChannel4ProtocolCrypt::writeProtocolCrypt durchgefuehrt

    // Zu 10: Packetdaten ggf. verschluesseln
    if ( (err = writeProtocolCrypt()) != IV_ERROR_SUCCESS )
    {	
        trace( "%s(%i):ERROR(1): %s() [ writeProtocolCrypt result: %i ]\n", __FILE__, __LINE__, __func__, err );
        // Ggf. Packetdaten raustracen.
        if ( isTrace() >= 2 )
            tracePacket( _packetWrite );

        return err;
    }
    // 11.
    // Wird von Transmit-Funktion erledigt

    trace( "%s(%i):INFO(1): %s() (id: %i, %i -> %i, X: %02X, B: %02X, M: %i, F: %i, PacketNummer: %i)\n", __FILE__, __LINE__, __func__, getId(), getDeviceId(), getDeviceIdRemote(),
        _packetWrite[6], _packetWrite[IV_PACKET_LENGTH_HEAD + _packetWrite[3]], module, func, _packetWrite[IV_PACKET_LENGTH_HEAD] );

    // Protokoll-Paket senden
    if ( (err = writeProtocol( _packetWrite, writeTrials )) != IV_ERROR_SUCCESS )
    {	
        trace( "%s(%i):ERROR(1): %s() [ writeProtocol result: %i ]\n", __FILE__, __LINE__, __func__, err );
        return err;
    }

    // Zeitpunkt des Packetversands halten
    _timeTransmitPacket = time( NULL );

    // Paketanzahl und Nutzdatengroesse fuer Statistik mitprotokollieren
    _packetWriteCounter++;
    _packetWritePayloadSize += count;

    return err;
}

// ##################################################################
/**
* @brief Lesen eines Packets inklusive Pruefung von Kopf und Gesamt-CRC.
*/
int Channel4Protocol::readProtocol( unsigned char *buffer, int readTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %i )\n", __FILE__, __LINE__, __func__, buffer, readTrials );

    int countEscape = 0;
    int received;
    int escape = FALSE;//, escapeError = FALSE;

    _clockRead = 0;
    clock_t clockRx = 0;

    int res;
    int error = IV_ERROR_SUCCESS;

    int receiveCount = IV_PACKET_LENGTH_HEAD;
    int receivedCount = 0;

    _skipCount = 0;

    // Liegt Paketstart breits vor?
    if ( _packetStart )
    {
        buffer[0] = IV_PACKET_SIGN_START;
        receivedCount++;
        // bereits gelesenen Paketstart uebernommen
        _packetStart = FALSE;
    }

    // Datenstrom zeichenweise lesen und Paketstartzeichen herausfiltern
    for ( ; receivedCount < receiveCount; ++receivedCount ) 
    {
        // Einstieg zum lesen eines Zeichens
        received = 0;

        // Empfangszeit fuer Pakete ermitteln.
        clock_t clockStart = clock();
        if ( (res = readData( &buffer[receivedCount], 1, received, readTrials )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s() [ ReceiveData fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
            return (error != IV_ERROR_SUCCESS) ? error : res;
        }
        clockRx += clock() - clockStart;

        // Empfangene Anzahl Zeichen erhoehen
        _amountSize += received;

        // Error abarbeiten
        if ( error != IV_ERROR_SUCCESS ) 
        {
            // Erstes Zeichen muss Startzeichen sein.
            if ( buffer[receivedCount] != IV_PACKET_SIGN_START ) 
            {
                // Startzeichen fehlt noch
                trace( "%s(%i):WARNING(1): %s() [ Skip first Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, buffer[receivedCount] );
                // Maximal zu lesende Zeichen ohne Paketstart
                if (++_skipCount > IV_SKIPCOUNT) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Maximal skip count reached: %i ]\n", __FILE__, __LINE__, __func__, error );
                    return -IV_ERROR_SKIPCOUNT;
                }

                // Emfpang zuruecksetzen
                receivedCount = -1;
                continue;
            } 
            else 
            {
                // Packetstart gelesen
                _packetStart = TRUE;
            }

            trace( "%s(%i):ERROR(1): %s() [ error: %i ]\n", __FILE__, __LINE__, __func__, error );
            return error;
        }

        // Erstes Zeichen muss Startzeichen sein.
        if ( !receivedCount && buffer[receivedCount] != IV_PACKET_SIGN_START ) 
        {	
            // Startzeichen fehlt noch
            trace( "%s(%i):WARNING(1): %s() [ Skip first Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, buffer[receivedCount] );

            // Paketzeit reset
            //ct1 = clock(); 

            // Maximal zu lesende Zeichen ohne Paketstart
            if ( ++_skipCount > IV_SKIPCOUNT ) 
            {
                trace("%s(%i):ERROR(1): %s() [ Maximal skip count reached: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_SKIPCOUNT);
                return -IV_ERROR_SKIPCOUNT;
            }

            // Empfang zuruecksetzen
            receivedCount = -1;
            continue;
        }
        else if ( receivedCount && buffer[receivedCount] == IV_PACKET_SIGN_START ) 
        {
            // Ein Startzeichen wurde im Paket empfangen.
            trace("%s(%i):ERROR(1): %s() [ Startbyte in packet: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_START_IN_PACKET);

            _packetStart = TRUE;
            return -IV_ERROR_START_IN_PACKET;
        } 
        else if ( receivedCount && (buffer[receivedCount] == IV_PACKET_SIGN_ESCAPE) ) 
        {
            // Escape bereits eingeschaltet?
            if ( escape ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ Double escape: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_ESCAPE );

                // Escape ausschalten
                escape = FALSE;
                // Fehler setzen
                error = -IV_ERROR_ESCAPE;
                _skipCount = 0;
                // Emfpang zuruecksetzen
                receivedCount = -1;
                continue;
            }

            // Escape einschalten
            escape = TRUE;
            // Emfpangszeiger zuruecksetzen
            receivedCount--;
            continue;
        } 
        else if ( escape ) 
        {
            // Escape ausschalten
            escape = FALSE;

            // Zu escapierendes Zeichen
            switch ( buffer[receivedCount] ) 
            {
            case 0:
                buffer[receivedCount] = IV_PACKET_SIGN_ESCAPE;
                break;

            case 1:
                buffer[receivedCount] = IV_PACKET_SIGN_START;
                break;

            case 2:
                // Bleibt im Puffer erhalten
                break;

            default:
                trace( "%s(%i):ERROR(1): %s() [ Unknown escape-sequenz: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_ESCAPESEQUENZ );
                //escapeError =  -IV_ERROR_ESCAPESEQUENZ;
                _skipCount = 0;
                // Emfpang zuruecksetzen
                receivedCount = -1;
                continue;
            }

            // Escapezeichen mitzaehlen
            countEscape++;
        }

        // Paketdaten auswerten
        // Es gibt 2 Auswertesituationen.
        // 1. Der Paketkopf wurde komplett empfangen -> Kopf-CRC pruefen. Geamtanzahl Packetdaten setzen.
        // 2. Das gesamte Paket wurde empfangen. -> Gesamt-CRC pruefen
        if ( (receivedCount + 1) == receiveCount ) 
        {
            // Paketkopf auswerten
            if ( (receivedCount + 1) == IV_PACKET_LENGTH_HEAD ) 
            {
                // CRC pruefen
                if ( crc16( buffer, receivedCount + 1 ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad head-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_HEAD );

                    int p = 0; // Fehlerhafte Kopfdaten ausgeben
                    for (; p < (receivedCount + 1); p++) 
                    {
                        // Fehlerhafte Kopfdaten trace
                        trace( "%s(%i):WARNING(1): %s() [ Skip( %3u ) head Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, p + 1, buffer[p] );
                    }

                    error = -IV_ERROR_CRC_HEAD;
                    // Emfpang zuruecksetzen
                    receivedCount = -1;
                    continue;
                }

                // Paketgroesenangabe pruefen
                if ( *reinterpret_cast<unsigned short *>( &buffer[4] ) < IV_PACKET_LENGTH_FRAME
                    || *reinterpret_cast<unsigned short *>( &buffer[4] ) >= IV_PACKET_LENGTH_MAX )
                {
                    error = -IV_ERROR_PACKETLENGTH;
                    // Emfpang zuruecksetzen
                    receivedCount = -1;
                    continue;
                }

                // Neue Empfangsgroesse setzen
                receiveCount = *reinterpret_cast<unsigned short *>( &buffer[4] );
            }
            else // Gesamt-CRC pruefen
            {
                // CRC pruefen
                if ( crc16( buffer, receivedCount + 1 ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad packet-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_PACKET );

                    int p = 0; // Fehlerhafte Kopf- Paketdaten ausgeben
                    for ( ; p < (receivedCount + 1); ++p ) 
                    {
                        // Fehlerhafte Kopfdaten trace
                        trace( "%s(%i):WARNING(1): %s() [ Skip( %3u ) %s Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, p + 1, (p < IV_PACKET_LENGTH_HEAD) ? "head" : "paket", buffer[p] );
                    }

                    error = -IV_ERROR_CRC_PACKET;
                    // Emfpang zuruecksetzen
                    receivedCount = -1;
                    continue;
                }
            }
        } 
    }

    // Gesamtzeit des Lesevorgangs fuer Statistik halten
    _clockRead = clockRx;
    _clocksRead += _clockRead;

    // Ggf. Packetdaten raustracen.
    if ( isTrace() >= 2 )
    {
        trace( "%s(%i):INFO(1): %s() ", __FILE__, __LINE__, __func__ );
        tracePacket( buffer );
    }

    return IV_ERROR_SUCCESS;
}

// ##################################################################
/**
* @brief Lesen eines Packets inklusive Pruefung von Kopf und Gesamt-CRC.
*/
int Channel4Protocol::readTimeboyStation( unsigned char *buffer, int readTrials )
{
    trace( "%s(%i):INFO(3): %s( %p, %i )\n", __FILE__, __LINE__, __func__, buffer, readTrials );

    static unsigned int crcLast = 0;
    static int crcRepeated = 0;
    int received;

    _clockRead = 0;
    clock_t clockStart = clock();
    clock_t clockEnd = clockStart + _timeout;
    clock_t clockEndActive = clockStart + IV_PENDING_TIMEOUT;

    int res = IV_ERROR_SUCCESS;

    int toReceive = IV_PACKET_LENGTH_FRAME; // Haelt die zu empfangende Anzahl Bytes
    int receivedCount = 0;                  // Haelt die Gesamtanzahl gelesener Bytes in _tmp
    int headSearches = 0;                   // Haelt die Anzahl der Leseversuche eines Packetanfanges beim Timeboy

    int receivePos = 0; // Position des Tmp - Lesepuffers
    int packetPos = 0;  // Position des Paketpuffers

    // Enum fuer Kommunikationsstatus.
    enum 
    {
        begin,         // Beginn der Paketauswertung.
        headSearch,    // Suche des Paketbeginns '~'
        headReceive,   // Daten des Paketkopfes werden empfangen.
        headComplete,  // Paketkopf komplett, CRC-Pruefung war OK.
        bodyReceive,   // Paketdaten werden empfangen.
        complete,      // Paket komplett empfangen
        failed
    } receiveStatus; // Aktueller Empfangsstatus

    // Aktuellen Empfangsstatus setzen.
    receiveStatus = begin;

    // Solange Zeichen zu empfangen sind.
    for ( ; receiveStatus != complete && receiveStatus != failed; )
    {

        // Je nach aktuellem Empfangsstatus.
        switch ( receiveStatus )
        {
        case begin: // Beginn der Paketauswertung.

            // Temp - Lesepuffer zuruecksetzen
            receivePos = 0;
            receivedCount = 0;
            // Auswertungspuffer zuruecksetzen
            packetPos = 0;

            // Zuerst zu lesende Menge auf Maximale Puffergroesse setzen.
            // Die Dockingstation liest an der Folgeadresse das naechste Datenpaket aus dem Timeboy!
            // Wir hier eine kleine Paketgroesse angegeben werden alle Folgepakete mit dieser Laenge geliefert.
            toReceive = MB_SLOT_BUFFER_DATA_SIZE_SMALL;//GetDockingBufferSize(getDeviceIdRemote());

            // Empfangsstatus wechseln
            receiveStatus = headSearch;
            break;

        case headSearch:

            // Erstes Datenbyte uebernehmen
            buffer[0] = _tmp[0];

            // Erstes Zeichen muss Startzeichen sein.
            if (buffer[0] != IV_PACKET_SIGN_START)
            {	
                headSearches++;
                trace( "%s(%i):WARNING(1): %s() [ Paket start wait (%i): 0x%02X ]\n", __FILE__, __LINE__, __func__, headSearches, buffer[0] );

                // Abbruch nach mehrfachem Nachlesen.
                if ( _timeboyBaseAddress == 0x70000000 && headSearches > 12 && clock() > clockEnd ) // 10 HeadSearches entspricht ca. 2500ms
                {
                    // Zu viele Versuche,
                    res = -IV_ERROR_SKIPCOUNT;
                    // Fehlerzustand annehmen.
                    receiveStatus = failed;
                    continue;
                }
                else if ( _timeboyBaseAddress == 0x70010000 && headSearches > 50 && clock() > clockEndActive )
                {
                    // Zu viele Versuche,
                    res = -IV_ERROR_SKIPCOUNT;
                    // Fehlerzustand annehmen.
                    receiveStatus = failed;
                    continue;
                }

                // Zeit fuer Versuche abgelaufen
                //if (clock() > clockEnd)
                //{
                //	// Zu viele Versuche,
                //	res = -IV_ERROR_RECEIVE;
                //	// Fehlerzustand annehmen.
                //	receiveStatus = failed;
                //	continue;
                //}

                // Erneut den Paketstart lesen.
                receiveStatus = begin;
                continue;
            }

            // Paketanfang erfolgt
            receiveStatus = headReceive;
            continue;

        case headReceive:

            packetPos++;
            receivePos++;
            // Naechsten Wert fuer Auswertung uebernehmen
            buffer[packetPos] = _tmp[receivePos];

            // Paketdaten auswerten
            // Es gibt 2 Auswertesituationen.
            // 1. Der Paketkopf wurde komplett empfangen -> Kopf-CRC pruefen. Geamtanzahl Packetdaten setzen.
            // 2. Das gesamte Paket wurde empfangen. -> Gesamt-CRC pruefen

            // Paketkopf auswerten
            if ( (packetPos + 1) == IV_PACKET_LENGTH_HEAD ) 
            {
                // CRC pruefen
                if ( crc16( buffer, IV_PACKET_LENGTH_HEAD ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad head-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_HEAD );

                    int p = 0; // Fehlerhafte Kopfdaten ausgeben
                    for (; p < packetPos; p++) 
                    {
                        // Fehlerhafte Kopfdaten trace
                        trace( "%s(%i):WARNING(1): %s() [ Skip head checked Databyte 0x%02X ]\n", __FILE__, __LINE__, __func__, buffer[p] );
                    }

                    res = -IV_ERROR_CRC_HEAD;
                    // Schnittstelle bereinigen
                    receiveStatus = failed;
                    continue;
                }

                // Paketgroesenangabe pruefen
                if ( *reinterpret_cast<unsigned short *>( &buffer[4] ) < IV_PACKET_LENGTH_FRAME
                    || *reinterpret_cast<unsigned short *>( &buffer[4] ) >= IV_PACKET_LENGTH_MAX )
                {
                    res = -IV_ERROR_PACKETLENGTH;
                    // Schnittstelle bereinigen
                    receiveStatus = failed;
                    continue;
                }

                // aktueller Status ist Paketkopf gueltig empfangen.
                receiveStatus = headComplete;
                continue;
            }
            break;

        case headComplete:
            // Anzahl nachzulesender Datenbytes setzen?
            if ( *reinterpret_cast<unsigned short *>( &buffer[4] ) > receivedCount )
            {
                // Neue Empfangsgroesse setzen
                toReceive = *reinterpret_cast<unsigned short *>( &buffer[4] );// - receivedCount;
            }

            // Empfangsstatus anpassen
            receiveStatus = bodyReceive;
            break;

        case bodyReceive:

            packetPos++;
            receivePos++;
            // Naechsten Wert fuer Auswertung uebernehmen
            buffer[packetPos] = _tmp[receivePos];

            if ( (packetPos + 1) == *reinterpret_cast<unsigned short *>( &buffer[4] ) ) 
            {		
                // Gesamt - CRC pruefen
                if ( crc16( buffer, packetPos + 1 ) ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Bad packet-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_PACKET );
                    res = -IV_ERROR_CRC_PACKET;
                    // Emfpang zuruecksetzen
                    receiveStatus = failed;
                    continue;
                }

                if ( crcLast != crc32( buffer, packetPos + 1 ) )
                {
                    // Letzten CRC von Packet merken.
                    crcLast = crc32( buffer, packetPos + 1 );
                    crcRepeated = 0;
                }
                else
                {
                    // Wiederholt gleiches Paket gelesen
                    if ( crcRepeated++ > 3 )
                    {
                        crcLast = 0;
                        crcRepeated = 0;
                        // Dreimalig gleiches Paket gelesen.
                        trace( "%s(%i):ERROR(1): %s() [ Repeated packet-crc: %i ]\n", __FILE__, __LINE__, __func__, -IV_ERROR_CRC_PACKET );
                        res = -IV_ERROR_SKIPCOUNT;
                        // Emfpang zuruecksetzen
                        receiveStatus = failed;
                        continue;
                    }

                }
                // Paket komplett empfangen
                receiveStatus = complete;
                continue;
            } 
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ receiveStatus unbekannt: %i ]\n", __FILE__, __LINE__, __func__, receiveStatus );
        }

        // Sind Zeichen zu lesen?
        if ( toReceive > 0 )
        {
            // Einstieg zum lesen eines Teilpaketes
            received = 0;

            if ( (res = _timeboyStation.readBytes( getDeviceIdRemote(), _timeboyBaseAddress, receivedCount, _tmp, toReceive )) != IV_ERROR_SUCCESS )
            {
                trace( "%s(%i):ERROR(1): %s() [ readBytes fail with: %i ]\n", __FILE__, __LINE__, __func__, res );
                return res;
            }

            // Empfangsanzahl angleichen
            received = toReceive;

            // Gesamtanzahl gelesener Bytes halten
            receivedCount += received;

            // Anzahl zu lesender Bytes wurde erreicht, somit ist diese nun 0
            toReceive -= received;

            // Statistik: Empfangene Anzahl Zeichen erhoehen
            _amountSize += received;
        }
    }

    // Gesamtzeit des Lesevorgangs fuer Statistik halten
    _clockRead = clock() - clockStart;
    _clocksRead += _clockRead;

    // Ggf. Packetdaten raustracen.
    if ( isTrace() >= 2 )
    {
        trace( "%s(%i):INFO(1): %s() ", __FILE__, __LINE__, __func__ );
        tracePacket( buffer );
    }

    return res;
}

// ##################################################################
/**
* @brief Lesen eines Packets aus der Kommunikationsschnittstelle.
*
* @param deviceIdRemote Anzusprechende Geraetenummer.
*/
// Empfangen und verarbeiten eines Pakets
//  deviceIdRemote -> Adresse (Geraetenummer) des Empfaengers
//  type     -> Pakettyp (Anfrage/Antwort +? Error)
//  module   -> Nummer des anzusprechenden Modules
//  func -> Nummer der Funktion im Module
//	buf	     -> Buffer fuer Funktionsparameter.
//	count    -> Anzahl der Funktionsparameter in Byte. Rueckgabewert bei Parameterempfang.
//
//  return -> <0 -> Kommunikationsfehler
//             0 -> senden und empfangen und ausfuehren erfolgreich
//			  >0 -> module-/Funktionsfehler
int Channel4Protocol::readPacket( short deviceId, unsigned char &type, short &module, short &func, unsigned char *buf, int &count, int readTrials )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i, %p, %i ]\n", __FILE__, __LINE__, __func__, deviceId, module, func, buf, count );

    int err = IV_ERROR_SUCCESS;

    // Adresse setzen
    setDeviceId( deviceId );

    /* 
    1.	(  8 Bit): Startzeichen ("~", 0x7E)
    2.	(  8 Bit): Empfängeradresse
    3.	(  8 Bit): Absenderadresse
    4.	(  8 Bit): Steuerinformationslänge
    5.	( 16 Bit): Paketgesamtlänge
    6.	(  8 Bit): Flags - für Verschlüsselung und Flashprüfung (z. B. für aktive Verschlüsselung)
    7.	(  8 Bit): Reserve
    8.	( 16 Bit): CRC über die ersten 7 Felder
    9.	(n*8 Bit): Steuerinformationsdaten (in Bezug auf Feld 5)
    10.	(n*8 Bit): Zu transportierende Daten
    11.	( 16 Bit): CRC über Felder 1 - 10.
    */

receive_again:
    // Wird ueber Dockingstation kommuniziert?
    if ( isTimeboySlot( getDeviceIdRemote() ) )
    {
        // Protokoll-Paket empfangen
        if ( (err = readTimeboyStation( _packetRead, readTrials )) != IV_ERROR_SUCCESS )
        {
            trace( "%s(%i):ERROR(1): %s() [ readTimeboyStation error: %i ]\n", __FILE__, __LINE__, __func__, err );
            return err;
        }
    }
    else
    {
        // Protokoll-Paket empfangen
        if ( (err = readProtocol( _packetRead, readTrials )) != IV_ERROR_SUCCESS ) 
        {
            trace( "%s(%i):ERROR(1): %s() [ readProtocol error: %i ]\n", __FILE__, __LINE__, __func__, err );
            return err;
        }
    }
  
    //	Empfaengeradresse pruefen
    if ( IV_PACKET_DEVICEID(_packetRead) != getDeviceId() ) 
    {
        err = -IV_ERROR_ADDRESS;
        trace( "%s(%i):WARNING(1): %s()  [ Wrong deviceId %i for %i ]\n", __FILE__, __LINE__, __func__, IV_PACKET_DEVICEID(_packetRead), getDeviceId() );

        // Handelt es sich um keine Docking,
        // weil dort kein Datenstromabriss erfolgt, wuerde eine Entlosschleife entstehen.
        if ( isTimeboySlot( getDeviceIdRemote() ) )
        {
            return err;
        }

        goto receive_again;
    }

    // Variable für die Flags in Member halten
    _lastFlags = IV_PACKET_FLAGS( _packetRead );
    /////////////////////////////////
    // Die Flags der Flashprüfen prüfen
    if ( _lastFlags & IV_PACKET_FLAG_FLASH_SERVICE )
    {	// Flash Service: Globaler Wear-Level ist >= 6
        trace( "%s(%i):WARNING(1): %s() [ Flash Service Flag is set 0x%02X ]\n", __FILE__, __LINE__, __func__, _lastFlags );
        // Flash Service occurce
    }
    if ( _lastFlags & IV_PACKET_FLAG_FLASH_WARNING )
    {	// Flash Warning: CRC Fehler im Ringspeicher oder Listenbereich. (Setup und Listen neu übertragen)
        trace( "%s(%i):WARNING(1): %s() [ Flash Warning Flag is set 0x%02X ]\n", __FILE__, __LINE__, __func__, _lastFlags );
        // Flash Warning occurce
    }
    if ( _lastFlags & IV_PACKET_FLAG_FLASH_LOCKED )
    {	// Flash Error: Keine freien Seiten mehr verfügbar, Gerät befindet sich im Bedienstopp.
        trace( "%s(%i):WARNING(1): %s() [ Flash Error Flag is set 0x%02X ]\n", __FILE__, __LINE__, __func__, _lastFlags );
        // Flash Error occurce
    }

    // Zu Punkt 9: Steuerinformation Packetzaehler
    // Ist Packet eine Antwort auf Anfrage?
    if ( _packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3]] & IV_PACKET_ANSWER ) 
    {
        // Steuerinformation verfuegbar?
        if ( _packetRead[3] > 0 ) 
        {
            // Ist Packetnummer der letzten Anfrage?
            // Oder handelt es sich um einen Fehler vom Typ "fehlerhafter gesamt-crc", "fehlerhafte Entschluesselung",
            // hierbei wird der Packetcounter auf 0 gesetzt das Modul und die Funktion ist 0xFFFF,
            // weil diese Daten nicht sichergestellt werden koennen.
            if ( _packetRead[IV_PACKET_LENGTH_HEAD] == _packetNumber
                || (_packetRead[IV_PACKET_LENGTH_HEAD] == 0 && _packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3]] & IV_PACKET_ERROR
                && *reinterpret_cast<unsigned short *>( &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char )] ) == 0xFFFF
                && *reinterpret_cast<unsigned short *>( &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char ) + sizeof( unsigned short )] ) == 0xFFFF) ) 
            {
                // Packetnummer erhoehen
                _packetNumber++;
            } 
            else 
            {
                err = -IV_ERROR_ADDRESS;
                trace( "%s(%i):WARNING(1): %s() [ Skip wrong packet number %i instead %i ]\n", __FILE__, __LINE__, __func__, _packetRead[IV_PACKET_LENGTH_HEAD], _packetNumber );
                // Ggf. Packetdaten raustracen.
                if ( isTrace() >= 2 )
                    tracePacket( _packetRead );

                // Handelt es sich um keine Docking,
                // weil dort kein Datenstromabriss erfolgt, wuerde eine Entlosschleife entstehen.
                if ( isTimeboySlot( getDeviceIdRemote() ) )
                {
                    return err;
                }
                goto receive_again;
            }
        }
    }

    // Empfangene Daten ggf. entschluesseln
    if ( (err = readProtocolCrypt()) != IV_ERROR_SUCCESS ) 
    {
        trace( "%s(%i):ERROR(1): %s() [ readProtocolCrypt error: %i ]\n", __FILE__, __LINE__, __func__, err );
        return err;
    }

    // Packettyp uebernehmen
    type = _packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3]];
    // Anzahl der Datenbyte
    count = *reinterpret_cast<unsigned short *>( &_packetRead[4] ) - IV_PACKET_LENGTH_FRAME - _packetRead[3] - sizeof( unsigned char ) - sizeof( unsigned short ) * 2;

    // Modul und Funktion uebernehmen
    module = *reinterpret_cast<unsigned short *>( &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char )] );
    func = *reinterpret_cast<unsigned short *>( &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char ) + sizeof( unsigned short )] );

    // Daten uebernehmen
    memcpy( buf, &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char ) + sizeof( unsigned short ) * 2], count );

    trace( "%s(%i):INFO(1): %s() (id: %i, %i <- %i, X: %02X, B: %02X, M: %i, F: %i, PacketNummer: %i)\n", __FILE__, __LINE__, __func__, getId(), getDeviceId(), _packetRead[2],
        _packetRead[6],
        _packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3]], 
        *reinterpret_cast<unsigned short *>( &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char )] ),
        *reinterpret_cast<unsigned short *>( &_packetRead[IV_PACKET_LENGTH_HEAD + _packetRead[3] + sizeof( unsigned char ) + sizeof( unsigned short )] ),
        (_packetRead[3]) ? _packetRead[IV_PACKET_LENGTH_HEAD] : 0 );

    // Zeitpunkt des Packetempfangs halten
    _timeReceivePacket = time( NULL );

    // Paketanzahl und Nutzdatengroesse fuer Statistik mitprotokollieren
    _packetReadCounter++;
    _packetReadPayloadSize += count;

    // remoteCall-Back Funktion aufrufen, ggf. abbruch.
    if ( _callBackFunction ) 
    {
        if ( !_callBackFunction() ) 
        {	
            trace( "%s(%i):ERROR(1): %s() [ CallBack abbruch ]\n", __FILE__, __LINE__, __func__ );
            return -IV_ERROR_BREAK;
        }
    }

    return err;
}

// ##################################################################
/**
* @brief Remoteaufruf einer lokalen Modulfunktion.
*/
void Channel4Protocol::tracePacket( unsigned char *packetBuf, bool crlf )
{
    int len = *reinterpret_cast<unsigned short *>( &packetBuf[4] );
    if ( len <= IV_PACKET_LENGTH_DATA )
    {
        char tmp[10];
        char *buf = new char[len * 3 + 10];
        memset( buf, 0, len * 3 + 10 );
        for ( len = 0; len < *reinterpret_cast<unsigned short *>( &packetBuf[4] ); ++len )
        {
            if ( len == 0 )
                sprintf( tmp, "[%02X ", packetBuf[len] );
            else if ( len == (*reinterpret_cast<unsigned short *>( &packetBuf[4] ) - 1) )
                sprintf( tmp, "%02X]%c", packetBuf[len], crlf ? '\n' : ' ' );
            else
                sprintf( tmp, "%02X ", packetBuf[len]);

            strcat( buf, tmp );
        }

        trace( buf );
        delete[] buf;
    }
}

// ##################################################################
/**
* @brief Remoteaufruf einer lokalen Modulfunktion.
*/
int Channel4Protocol::remoteCall( short module, short func, unsigned char *data, int &count )
{
    (void)module;
    (void)func;
    (void)data;
    (void)count;
    // Fehler: Das geforderte Modul wird nicht unterstuetzt.
    return 0x0100;
}

// ##################################################################
/**
* @brief Ausfuehrung einer Modulfunktion in ihrem jeweiligen Kontext.
*
*/
int Channel4Protocol::call( short context, short module, short func, unsigned char *buf, int &count, int contextValue )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i, %p, %i, %i )\n", __FILE__, __LINE__, __func__, context, module, func, buf, count, contextValue );

    int trials, receiveTrials, retries, continues = 0;
    int txCount = 0, rxCount;
    unsigned char txTmp[IV_PACKET_LENGTH_DATA], rxTmp[IV_PACKET_LENGTH_DATA];
    short rxModule, rxFunction;
    unsigned char type = 0;
    int err = IV_ERROR_SUCCESS;

    switch ( context )
    {
    case IV_CONTEXT_TRANSMIT_ANSWER: // Einfacher Versand eines Packetes
    case IV_CONTEXT_TRANSMIT_ANSWER_ERROR: // Einfacher Versand eines Packetes
        // Temppuffer fuer Transmit-Packet mit Daten fuellen.
        txCount = count;
        memcpy( txTmp, buf, txCount );
        retries = (contextValue == -1) ? IV_WRITE_TRIALS : contextValue;
        receiveTrials = 0;
        _timeboyBaseAddress = 0x70010000;
        break;

    case IV_CONTEXT_MASTER: // Ausführung (Versant und Empfang) eines Befehls
        // Temppuffer fuer Transmit-Packet, falls ein Request des Geraetes dazwischen kommt und
        // die selben Daten nochmals gesendet werden.
        txCount = count;
        memcpy( txTmp, buf, txCount );
        retries = (contextValue == -1) ? IV_WRITE_TRIALS : contextValue;
        receiveTrials = IV_READ_TRIALS;
        _timeboyBaseAddress = 0x70000000;
        break;

    case IV_CONTEXT_SLAVE: // Empfang und Ausführung eines Befehls
        retries = (contextValue == -1) ? IV_READ_TRIALS : contextValue; 
        receiveTrials = 1;
        _timeboyBaseAddress = 0x70010000;
        break;

    default:
        trace( "%s(%i):ERROR(1): %s() [ Unbekannter Context ]\n", __FILE__, __LINE__, __func__ );
        return -IV_ERROR_FAIL;
    }

    // Einen Versuch sicherstellen
    if ( retries <= 0 )
        retries = 1;

    for ( trials = 0; trials < retries && continues < 10; ++trials ) 
    {
        // Ist ein Kommunikationsbefehl abzuarbeiten?
        if ( context == IV_CONTEXT_MASTER || context == IV_CONTEXT_TRANSMIT_ANSWER || context == IV_CONTEXT_TRANSMIT_ANSWER_ERROR ) 
        {
            unsigned char type = (context == IV_CONTEXT_MASTER) ? IV_PACKET_REQUEST : IV_PACKET_ANSWER;
            if ( context == IV_CONTEXT_TRANSMIT_ANSWER_ERROR )
                type |= IV_PACKET_ERROR;

            // Befehlspacket senden 
            if ( (err = writePacket( getDeviceIdRemote(), type, module, func, txTmp, txCount )) != IV_ERROR_SUCCESS ) 
            {
                trace( "%s(%i):ERROR(1): %s() [ writePacket context: %i, error: %i ]\n", __FILE__, __LINE__, __func__, context, err );
                break;
            }

            // Kein lesen erforderlich?
            if ( context != IV_CONTEXT_MASTER )
                break;
        } 

receive_again:
        // Befehlspacket empfangen
        if ( (err = readPacket( getDeviceId(), type, rxModule, rxFunction, rxTmp, rxCount, receiveTrials )) != IV_ERROR_SUCCESS ) 
        {
            trace( "%s(%i):ERROR(1): %s() [ readPacket error: %i ]\n", __FILE__, __LINE__, __func__, err );
            if (err == -IV_ERROR_BREAK)
                break; // Abbruch durch remoteCall-Back (Anwender)

            if ( context == IV_CONTEXT_MASTER )
            {
                // Anzahl unbeantworteter Pakete fuer Statistik halten
                _packetLostCounter++;
            }
            else if ( context == IV_CONTEXT_SLAVE )
            {
                switch ( err )
                {
                case IV_ERROR_CRC_PACKET: // Gesamt-CRC des Packets ist falsch.
                    break;

                case IV_ERROR_PACKETLENGTH: // Gesamt-Laenge des Packets ist falsch.
                    break;
                }
            }

            if ( err == -IV_ERROR_CHANNEL_OPEN_REMOTE ) // Remote hat die Verbindung geschlossen
                break; 

            continue;
        }

        // Packetbefehl auswerten
        switch ( type ) 
        {
        case (IV_PACKET_VERSION_4 | IV_PACKET_REQUEST):
            {

                short deviceIdRemote = setDeviceIdRemote( _packetRead[2] );

                // BUGFIX
                // Der Bootloader liefert als aktive Antwort auf IV_FUNCTION_BEGIN_SET_SEGMENT anstatt
                // IV_FUNCTION_BEGIN_SET_SEGMENT_OK ebenfalls den Wert IV_FUNCTION_BEGIN_SET_SEGMENT
                // Ist aktuell Warten auf diese Aktive Antwort?
                if ( context == IV_CONTEXT_SLAVE && module == 0x0001/*IV_MODULE_SYSTEM*/ && func == 0x0010/*IV_FUNCTION_BEGIN_SET_SEGMENT_OK*/ ) 
                {
                    // Korrektur des Funktionswertes, wenn nötig
                    rxFunction = func;
                }

                // Funktion der entsprechenden Modulklasse aufrufen.
                // Die Routinen senden die Antwort selber, nur Fehler werden nachfolgend fuer
                // alle gemeinsam abgehandelt.
                if ( (err = remoteCall( rxModule, rxFunction, rxTmp, rxCount )) != IV_ERROR_SUCCESS ) 
                {	
                    trace( "%s(%i):ERROR(1): %s() [ remoteCall( %i, %i, %p, %i ) ]\n", __FILE__, __LINE__, __func__, rxModule, rxFunction, rxTmp, rxCount );

                    // Fehler als Antwort senden.
                    *(short*)rxTmp = (short)err;
                    if ( (err = writePacket( getDeviceIdRemote(), IV_PACKET_ANSWER | IV_PACKET_ERROR, rxModule, rxFunction, rxTmp, sizeof(short) )) != IV_ERROR_SUCCESS ) 
                    {	
                        trace( "%s(%i):ERROR(1): %s() [ writePacket error: %i ]\n", __FILE__, __LINE__, __func__, err );
                        setDeviceIdRemote( deviceIdRemote );
                        return err;
                    }
                }

                // Auf Vorgaenger wechseln.
                setDeviceIdRemote( deviceIdRemote );

                // Aktive Antwort OK?
                if ( (context == IV_CONTEXT_SLAVE) && (module != IV_MODULE_BASE && rxModule == module && rxFunction == func) ) 
                {
                    // Datenkopie uebernehmen
                    memcpy( buf, rxTmp, rxCount );
                    count = rxCount;
                    err = IV_ERROR_SUCCESS;
                } 
                else if ( (context == IV_CONTEXT_MASTER) || (context == IV_CONTEXT_SLAVE && module != IV_MODULE_BASE) )
                {
                    // Wurde Befehl gesendet und die Antwort ist noch zu lesen,
                    // oder wird noch auf die richtige aktive Antwort gewartet.
                    trace( "%s(%i):INFO(1): %s() [ Receive again ]\n", __FILE__, __LINE__, __func__ );
                    goto receive_again;
                }
                break;
            }

        case (IV_PACKET_VERSION_4 | IV_PACKET_REQUEST | IV_PACKET_ERROR):
            {

                if ( rxCount != sizeof( short ) ) 
                {
                    // Falsche Parameteranzahl
                    trace( "%s(%i):ERROR(1): %s() [ Fehlerrequest: falsche Parameteranzahl ]\n", __FILE__, __LINE__, __func__ );
                }

                // Fehlercode zuweisen
                short e = *reinterpret_cast<short *>( rxTmp );

                // Fehler beantworten.
                if ( (err = writePacket( _packetRead[2], IV_PACKET_ANSWER | IV_PACKET_ERROR, rxModule, rxFunction, _packetData, sizeof( short ) )) != IV_ERROR_SUCCESS ) 
                {	
                    trace( "%s(%i):ERROR(1): %s() [ writePacket error: %i ]\n", __FILE__, __LINE__, __func__, err );
                    return err;
                }

                // Aktive Antwort OK?
                if ( (context == IV_CONTEXT_SLAVE) && (module != IV_MODULE_BASE && rxModule == module && rxFunction == func) ) 
                {
                    // Nur als Aktive Antwort sinnvoll!!!
                    return e;
                } 
                else 
                {
                    // Unerwartete Anfrage erhalten
                    trace( "%s(%i):ERROR(1): %s() [ Fehlerrequest: Unerwartete Anfrage ]\n", __FILE__, __LINE__, __func__ );
                }
                break;
            }

        case (IV_PACKET_VERSION_4 | IV_PACKET_ANSWER):

            // Antwort auf zuvor gesendeten Befehl?
            if ( context == IV_CONTEXT_MASTER ) 
            {
                // Antwort von gefordertem Absender?
                if ( IV_PACKET_DEVICEID_REMOTE(_packetRead) != getDeviceIdRemote() ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Falscher Absender ]\n", __FILE__, __LINE__, __func__ );
                    return -IV_ERROR_ADDRESS;
                }

                // Antwort auf aktuelles Modul/Function
                if ( rxModule != module || rxFunction != func ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Falsche Module/Function ]\n", __FILE__, __LINE__, __func__ );
                    goto receive_again;
                }

                // Daten in Buffer der Applikation kopieren
                memcpy( buf, rxTmp, rxCount );
                count = rxCount;
                err = IV_ERROR_SUCCESS;
            } 
            else 
            {
                // Unerwartete Antwort erhalten
                trace( "%s(%i):ERROR(1): %s() [ Unerwartete Antwort ]\n", __FILE__, __LINE__, __func__ );
            }
            break;

        case (IV_PACKET_VERSION_4 | IV_PACKET_ERROR):
        case (IV_PACKET_VERSION_4 | IV_PACKET_ANSWER | IV_PACKET_ERROR):

            // Antwort auf zuvor gesendeten Befehl?
            if ( context == IV_CONTEXT_MASTER ) 
            {
                // Antwort von gefordertem Absender?
                if ( IV_PACKET_DEVICEID_REMOTE(_packetRead) != getDeviceIdRemote() ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Falscher Absender ]\n", __FILE__, __LINE__, __func__ );
                    return -IV_ERROR_ADDRESS;
                }

                // Wenn der Gesamt-CRC fehlerhaft ist oder die Entschluesselung nicht durchgefuehrt werden konnte,
                // dann wird in der Antwort Modul und Funktion auf -1 gesetzt.
                if ( static_cast<unsigned short>( rxModule ) == 0xFFFF && static_cast<unsigned short>( rxFunction ) == 0xFFFF )
                {

                }
                else if ( rxModule != module || rxFunction != func )
                { 
                    // Antwort auf aktuelles Modul/Function
                    trace( "%s(%i):ERROR(1): %s() [ Falsche Module/Function ]\n", __FILE__, __LINE__, __func__ );
                    goto receive_again;
                }

                // Fehlercode uebernehmen
                err = *reinterpret_cast<short *>( rxTmp );

                // Pruefen ob Modulfehler vorliegt und die Anfrage erneut gesendet werden muss.
                // PROT_MODULE_ERROR_CRC                    (0x0700) // Gesamt CRC von Packet ist falsch
                // PROT_MODULE_ERROR_CONTROL                (0x0800) // Zu wenig Steuerinformationen
                // PROT_MODULE_ERROR_PROTOCOL               (0x0900) // Protokollversion passt nicht
                // PROT_MODULE_ERROR_NO_QUERY               (0x0A00) // Anfrage erwartet, Antwort erhalten
                // PROT_MODULE_ERROR_BUSY                   (0x0B00) // Antwort erwartet, Anfrage erhalten
                // PROT_MODULE_ERROR_SIZE                   (0x0C00) // Packetgroesse ueberschreitet reservierte Puffergroesse
                // PROT_MODULE_ERROR_DECRYPT                (0x0D00) // Entschluesselung des Packetinhaltes bring Fehler.
                // PROT_MODULE_ERROR_CRYPTED                (0x0E00) // Geraet arbeitet verschluesselt und Anfrage war unverschluesselt.

                if ( err == 0x0700 || err == 0x0B00 ) 
                {
                    trace( "%s(%i):ERROR(1): %s() [ Modulfehler %04X, wiederholtes senden. ]\n", __FILE__, __LINE__, __func__, err );
                    trials--; // Nochmaliges senden des Packets
                    continues++; // Abbruch nach n-maligem wiederholen.
                    continue;
                }
            } 
            else 
            {
                // Unerwartete Antwort erhalten
                trace( "%s(%i):ERROR(1): %s() [ Unerwartete Antwort. context: %i, modul: %i, func: %i ]\n", __FILE__, __LINE__, __func__, context, module, func );
            }
            break;

        default:
            trace( "%s(%i):ERROR(1): %s() [ Unbekannter Packettyp: %02X ]\n", __FILE__, __LINE__, __func__, type );
            goto receive_again;
        }

        // Schleife nach erfolgreicher Abarbeitung verlassen.
        break;
    }

    if ( context == IV_CONTEXT_MASTER ) 
    {
        double tm = static_cast<double>( (double)_clockWrite + (double)_clockRead ) / (double)CLOCKS_PER_SEC;
        //double clockTx = (double)_clockWrite / (double)CLOCKS_PER_SEC;
        //double clockRx = (double)_clockRead / (double)CLOCKS_PER_SEC;
        _qs = (double)((double)_clocksWrite + (double)_clocksRead) / (double)CLOCKS_PER_SEC;;

        if ( !(_packetReadCounter % 10) ) 
        {	
            double saveVal1 = (_writtenSize) ? ((double)(_writtenSize - _packetWritePayloadSize) / (double)_writtenSize) * 100 : 0;
            double saveVal2 = (_amountSize) ? ((double)(_amountSize - _packetReadPayloadSize) / (double)_amountSize) * 100 : 0;
            double saveVal3 = ((_writtenSize + _amountSize) == 0) ? 0 : double((_writtenSize - _packetWritePayloadSize) + (_amountSize - _packetReadPayloadSize)) / (double)(_writtenSize + _amountSize) * 100;
            double saveVal4 = (_packetReadCounter == 0) ? 0 : (double)_qs / (double)_packetReadCounter;

            //Trace("%s(%i):INFO(1): %s [ QS txBuf(%i ; %i / %i) [sendezeit+empfangszeit: %f][ empfangszeit: %f]{gesamt(empfangszeit/packetanzahl): %f} ]\n", __FILE__, __LINE__, __func__, (receivedCount+masked), (double)((double)ct2 - (double)m_ct) / (double)CLOCKS_PER_SEC, tm, (double)qs / (double)count);
            trace( "%s(%i):INFO(4): %s() [ txBuf: %i; %i/%i; %.2f; rxBuf: %i, %i; %i/%i; %.2f; s(txBuf/rxBuf): %i/%i; %.2f; t(txBuf+rxBuf): %f; t(rxBuf): %f; s(t(rxBuf+txBuf)/pa): %f; ]\n", __FILE__, __LINE__, __func__, 
                _packetWriteCounter, _writtenSize, _packetWritePayloadSize, saveVal1, 
                _packetReadCounter, _packetLostCounter, _amountSize, _packetReadPayloadSize, saveVal2, 
                _writtenSize + _amountSize, _packetWritePayloadSize + _packetReadPayloadSize, saveVal3,
                tm, (double)_clockRead / (double)CLOCKS_PER_SEC, saveVal4);
            /*
            double val1 = (m_transmittedSize == 0) ? 0 : ((double)(m_transmittedSize - m_transmittedSizeData) / (double)m_transmittedSize) * 100;
            double val2 = (m_receivedSize == 0) ? 0 : ((double)(m_receivedSize - _packetReadPayloadSize) / (double)m_receivedSize) * 100;
            double val3 = ((m_transmittedSize + m_receivedSize) == 0) ? 0 : double((m_transmittedSize - m_transmittedSizeData) + (m_receivedSize - _packetReadPayloadSize)) / (double)(m_transmittedSize + m_receivedSize) * 100;
            double val4 = (_packetReadCounter == 0) ? 0 : (double)_qs / (double)_packetReadCounter;

            Trace("%s(%i):INFO(4): %s() [ transmit { time %f, packetcount %i; size %i / datasize %i; overhead in prozent %.2f%% } ]\n", __FILE__, __LINE__, __func__, clockTx, _packetWriteCounter, m_transmittedSize, m_transmittedSizeData, val1);
            Trace("%s(%i):INFO(4): %s() [ receive { time %f, packetcount %i, packetcount unbeantwortet %i; size %i / datasize %i; overhead %.2f%% } ]\n", __FILE__, __LINE__, __func__, clockRx, _packetReadCounter, _packetLostCounter, m_receivedSize, _packetReadPayloadSize, val2);
            Trace("%s(%i):INFO(4): %s() [ summe(transmit + receive) { size %i (%.3fkByte) / datasize %i (%.3fkByte); overhead %.2f%% } ]\n", __FILE__, __LINE__, __func__, m_transmittedSize + m_receivedSize, (double)(m_transmittedSize + m_receivedSize) / (double)1024, m_transmittedSizeData + _packetReadPayloadSize, (double)(m_transmittedSizeData + _packetReadPayloadSize) / (double)1024, val3);
            Trace("%s(%i):INFO(4): %s() [ time summe(transmit + receive) { %f; durchschnittszeit pro einzelbefehl: %f } ]\n", __FILE__, __LINE__, __func__, tm, val4);
            */
        }
    }

    // Wurde die maximale Anzahl moeglicher Versuche erreicht?
    if ( trials == retries ) 
    {
        trace( "%s(%i):ERROR(1): %s() [ Maximalanzahl der Versuche erreicht. ]\n", __FILE__, __LINE__, __func__ );

        // Packetzaehler erhoehen!
        _packetNumber++;

        //		return -IV_ERROR_NO_REQUEST; // Die aktive Antwort kam nicht nach der vorgegebenen Dauer
    }

    return err;
}

//#########################################################
//####################### ERROR ###########################
//#########################################################

// ##################################################################
/**
* @brief Die gesamte Fehlerbehandlung zuruecksetzen.
*/
void Channel4Protocol::resetError()
{
    trace( "%s(%i):INFO(3): %s()\n", __FILE__, __LINE__, __func__ );

    // Fehlernummern loeschen
    _errorCommunicationNum = _errorModuleNum = _errorFunctionNum = 0;
    // Puffer die in Gebrauch waren, loeschen.
    if ( _errorCommunication.size() )
        _errorCommunication.clear();

    if ( _errorModule.size() )
        _errorModule.clear();

    if ( _errorFunction.size() )
        _errorFunction.clear();
}

// ##################################################################
/**
* @brief Logen der textuellen Fehlerentsprechung.
*
*/
void Channel4Protocol::traceError( short deviceIdRemote, short module, short func )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i )\n", __FILE__, __LINE__, __func__, deviceIdRemote, module, func );

    if ( _result < 0 )
    {		
        // Fehlernummern setzen
        _errorCommunicationNum = -_result;
        // Fehlertext ermitteln und eintragen.
        getErrorCommunication( _errorCommunicationNum, _errorCommunication );

        trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, -_result & 0xFFFF, _errorCommunication.c_str() );
    }
    else if ( _result > 0 )
    {
        // Fehlertext ermitteln und eintragen.
        if ( IV_RESULT_MODULE(_result) )
        {
            // Fehlernummern setzen
            _errorModuleNum = IV_RESULT_MODULE(_result);
            // Modulefehlertext ermitteln
            getErrorModule( module, _errorModuleNum, _errorModule );

            trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, _result & 0xFFFF, _errorModule.c_str() );
        }
        else
        {
            // Fehlernummern setzen
            _errorFunctionNum = IV_RESULT_FUNCTION(_result);
            // Funktionsfehlertext ermitteln
            getErrorFunction( module, func, _errorFunctionNum, _errorFunction );

            trace( "%s(%i):INFO(2): %s() [ %04X: %s ]\n", __FILE__, __LINE__, __func__, _result & 0xFFFF, _errorFunction.c_str() );
        }

    }
}

// ##################################################################
/**
* @brief Ermittlung des Fehlertextes zum Fehlercode fuer Kommunikationsfehler.
*/
void Channel4Protocol::getErrorCommunication( short error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %i, %p )\n", __FILE__, __LINE__, __func__, error, text.c_str() );

    switch ( error )
    {
    case IV_ERROR_SUCCESS:
        text = boost::str( boost::format( "( %i ): Kein Fehler" ) % error );
        break;

    case IV_ERROR_CHANNEL_OPEN:
        text = boost::str( boost::format( "( %i ): Schnittstelle geschlossen oder konnte nicht geöffnet werden." ) % error );
        break;

    case IV_ERROR_UNREACHABLE:
        text = boost::str( boost::format( "( %i ): Gerät antwortet nicht auf Anfrage." ) % error );
        break;

    case IV_ERROR_TRANSMIT:
        text = boost::str( boost::format( "( %i ): Es konnten keine Daten in die Schnittstelle geschrieben werden." ) % error );
        break;

    //case IV_ERROR_RECEIVE:
        //text = boost::str( boost::format( "( %i ): Es konnten keine Daten von der Schnittstelle gelesen werden." ) % error );
        //break;

    case IV_ERROR_FILE_OPEN:
        text = boost::str( boost::format( "( %i ): Datei konnte nicht geöffnet werden." ) % error );
        break;

    case IV_ERROR_CRC_PACKET:
        text = boost::str( boost::format( "( %i ): Prüfsumme (CRC) von Protokollpaket stimmt nicht." ) % error );
        break;

    case IV_ERROR_ADDRESS:
        text = boost::str( boost::format( "( %i ): Die Adressen (Sender/Empfänger) des Protokollpaketes sind fehlerhaft." ) % error );
        break;

    case IV_ERROR_COMMAND:
        text = boost::str( boost::format( "( %i ): Protokollbefehl der vierten Schicht ist falsch oder unbekannt." ) % error );
        break;

    //case IV_ERROR_PACKETLENGTH:
        //text = boost::str( boost::format( "( %i ): Die angegebene Paketlaenge stimmt nicht ueberein." ) % error );
        //break;

    case IV_ERROR_BREAK:
        text = boost::str( boost::format( "( %i ): Die Kommunikation wurde durch den Anwender (CallBack-Funktion) abgebrochen." ) % error );
        break;

    case IV_ERROR_FUNCTION_UNAVAILABLE:
        text = boost::str( boost::format( "( %i ): Funktion steht im vorliegenden Fall nicht zur Verfuegung." ) % error );
        break;

    case IV_ERROR_NEED_PASSWORD:
        text = boost::str( boost::format( "( %i ): Das angegebene Passwort stimmt nicht mit dem benötigten überein. Es wird ein Passwort gefordert." ) % error );
        break;

    case IV_ERROR_PASSWORD_LOCK:
        text = boost::str( boost::format( "( %i ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt." ) % error );
        break;

    case IV_ERROR_NEED_USER_PASSWORD:
        text = boost::str( boost::format( "( %i ): Fuer die verschluesselte Kommunikation mit dem Geraet wird ein Benutzer-Passwort benoetigt." ) % error );
        break;

    default:
        text = boost::str( boost::format( "( %i ): Nicht eingetragener Fehler!" ) % error );
        break;
    }
}

// ##################################################################
/**
* @brief Ermittlung des Fehlertextes zum Fehlercode fuer Modulfehler.
*/
void Channel4Protocol::getErrorModule( short module, short error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %p )\n", __FILE__, __LINE__, __func__, module, error, text.c_str() );

    // Pruefen ob Modulfehler vorliegt und die Anfrage erneut gesendet werden muss.
    switch ( error )
    {
    case 1:
        text = boost::str( boost::format( "( %i ): Das geforderte Modul wird nicht unterstützt." ) % error );
        break;

    case 2:
        text = boost::str( boost::format( "( %i ): Die geforderte Funktion wird nicht unterstützt." ) % error );
        break;

    case 3:
        text = boost::str( boost::format( "( %i ): Es wurden bei der Funktion zu wenige Paramteter angegeben." ) % error );
        break;

    case 4:
        text = boost::str( boost::format( "( %i ): Es wurden bei der Funktion zu viele Paramteter angegeben." ) % error );
        break;

    case 5:
        text = boost::str( boost::format( "( %i ): Geforderte Datenmenge wurde unterschritten." ) % error );
        break;

    case 6:
        text = boost::str( boost::format( "( %i ): Geforderte Datenmenge wurde überschritten." ) % error );
        break;

    case 7:
        text = boost::str( boost::format( "( %i ): Gesamt CRC des Packetes ist falsch.." ) % error );
        break;

    case 8:
        text = boost::str( boost::format( "( %i ): Es werden zusätzliche Steuerinformationen benötigt." ) % error );
        break;

    case 9:
        text = boost::str( boost::format( "( %i ): Die gesendete Protokollversion ist unbekannt." ) % error );
        break;

    case 10:
        text = boost::str( boost::format( "( %i ): Es wurde eine Anfrage erwartet, Antwort erhalten." ) % error );
        break;

    case 11:
        text = boost::str( boost::format( "( %i ): Es wurde eine Antwort erwartet, Anfrage erhalten." ) % error );
        break;

    case 12:
        text = boost::str( boost::format( "( %i ):  Packetgroesse und damit die reservierte Puffergroesse ueberschreitet." ) % error );
        break;

    case 13:
        text = boost::str( boost::format( "( %i ): Entschluesselung des Packetinhaltes bringt Fehler." ) % error );
        break;

    case 14:
        text = boost::str( boost::format( "( %i ): Geraet arbeitet verschluesselt und Anfrage war unverschluesselt." ) % error );
        break;

    case 15:
        text = boost::str( boost::format( "( %i ): Bei der Ausführung ist ein interner Timeout eingetreten." ) % error );
        break;

    case 16:
        text = boost::str( boost::format( "( %i ): Angesprochener Slot ist nicht vorhanden.") % error );
        break;

    case 17:
        text = boost::str( boost::format( "( %i ): Kein Timeboy im Slot vorhanden.") % error );
        break;

    case 18:
        text = boost::str( boost::format( "( %i ): Timeboytyp noch nicht erkannt.") % error );
        break;

    case 19:
        text = boost::str( boost::format( "( %i ): Es wird versucht ein TimeboyI-III anzusprechen.") % error );
        break;

    case 20: // Dockingstation liest Datensätze vom Timeboy oder überträgt Listen.
        text = boost::str( boost::format( "( %i ): Ausführung konnte wegen belegtem Kommunikationskanal nicht weitergereicht werden.") % error );
        break;


    case 28:
        text = boost::str( boost::format( "( %i ): Das angegebene Passwort stimmt nicht mit dem benötigten überein. Es wird ein Passwort gefordert.") % error );
        break;

    case 29:
        text = boost::str( boost::format( "( %i ): Gerät ist aufgrund von fehlerhaften Passworteingaben für 60 Sekunden gesperrt.") % error );
        break;


    default:
        text = boost::str( boost::format( "( %i ): Nicht abgefangener Modulfehler!" ) % error );
    }
}


// ##################################################################
/**
* @brief Ermittlung des Fehlertextes zum Fehlercode fuer Funktionsfehler.
*/
void Channel4Protocol::getErrorFunction( short module, short func, short error, std::string &text )
{
    trace( "%s(%i):INFO(3): %s( %i, %i, %i, %p )\n", __FILE__, __LINE__, __func__, module, func, error, text.c_str() );

    text = boost::str( boost::format( "( %i ): Nicht abgefangener Funktionsfehler!" ) % error );
}

// ##################################################################
/**
* @brief Workerthread fuer die Verarbeitung der Kommunikationsaufgaben.
*/
int Channel4Protocol::callServer()
{
    trace("%s(%i):INFO(3): %s\n", __FILE__, __LINE__, __func__);

#ifdef _MSC_VER
#ifndef MYCONST
#define MYCONST const
#endif
#else
#define MYCONST
#endif

    // Hauptschleife ausfuehren?
    _threadCommandServer = commandIdle;

    int res, failCount = 0, resumeCount = 0;
    SOCKADDR_IN sa;
    struct timeval seltime;
    fd_set readfds;
    int count;
    unsigned char data[IV_PACKET_LENGTH_DATA]; // 21 Byte Paketframe

    int len;

    // Timeout fuer Abfrage auf Daten
    seltime.tv_sec  = 0;
    seltime.tv_usec = 10000;

    // Fehlerzaehler neu initialisieren
    failCount = 0;

    switch (_threadCommandServer)
    {
    case commandIdle:
        {	
            // Socketdaten initialisieren
            FD_ZERO((fd_set FAR *) &readfds);
            FD_SET(_socket, (fd_set FAR *) &readfds);

            // Auf eintreffende Daten warten
            if ((res = select(FD_SETSIZE,
                (fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
                (MYCONST struct timeval FAR *)&seltime)) != SOCKET_ERROR)
            {
                // Stehen Daten an? (ansonsten Timeout)
                if (res)
                {					
                    // Paketabarbeitung einleiten
                    if ((res = call((short)IV_CONTEXT_SLAVE, (short)IV_MODULE_BASE, (short)IV_FUNCTION_BASE, data, count, 1)) != IV_ERROR_SUCCESS)
                    {						
                        if ((res == -IV_ERROR_CHANNEL_OPEN_REMOTE) || (failCount > 2))
                        {							
                            // Wenn Daten gemeldet werden,
                            // die Anzahl aber 0 ist, so wurde die Verbindung beendet.

                            // Hauptschleife verlassen
                            return IV_ERROR_FAIL;
                        }

                        // Fehlerzaehler erhoehen
                        failCount++;
                    }
                    else
                    {						
                        // Fehlerzaehler zuruecksetzen
                        failCount = 0;
                    }
                }
            }
            else
            {				
                trace("%s(%i):ERROR(1): %s [ select fail (%i) ]\n", __FILE__, __LINE__, __func__, WSAGetLastError());

                // Hauptschleife verlassen
                return IV_ERROR_FAIL;
            }
            break;}

    case commandSuspending:
        _threadCommandServer = commandSuspend;
        break;

    case commandSuspend: 
        // Kurze Auszeit.
        Sleep(10);
        break;

    case commandResuming:
        resumeCount = 0;
        _threadCommandServer = commandResume;
        break;

    case commandResume:

        resumeCount++;

        if (resumeCount >= 10)
        {
            // Sicherstellen, dass keine Anforderung auf Suspending besteht.
            if (_suspendCounterServer > 0)
            {
                _threadCommandServer = commandSuspend;
            }
            else
            {				
                _threadCommandServer = commandIdle;
            }
        }
        Sleep(30);
        break;
    }

    return IV_ERROR_SUCCESS;
}
