/**
* @file  FileWlanConfig.cpp
* @brief Implementierung der Klasse FileWlanConfig.
*/
//////////////////////////////////////////////////////////////////////

#include "filewlanconfig.h"
#include <memory.h>
#include <stdio.h>

FileWlanConfig::FileWlanConfig()
{
    // CRC32 Initialisierung
    InitCRC32Table();

    // Speicher reservieren.
    m_memWlanConfig = new unsigned char[MAX_MEM_WLANCONFIG];

    // Pruefsumme ob Dokumentdaten veraendert wurden.
    _crc = 0;

    // Dateiname
    memset(_fileName, 0, sizeof(_fileName));

    // Ausgangseinstellung
    resetError();
    reset();
}

FileWlanConfig::~FileWlanConfig()
{
    // Speicher freigeben.
    delete[] m_memWlanConfig;
}

void FileWlanConfig::reset()
{
    // Basis Reset
    FileIntelHex::reset();

    // Member initialisieren

    // Groessen setzen
    m_sizeData   = 0;

    // Pruefsumme zum Test auf Datenaenderung ruecksetzen.
    _crc = 0;

    // Speicher fuer WlanConfigDaten loeschen
    memset(m_memWlanConfig, 0x00, MAX_MEM_WLANCONFIG);

    // Hilfsstrukturen aufsetzen
    m_pWlanConfigNode   = (WLANCONFIGNODE*)m_memWlanConfig;
    m_pWlanConfigNode->version = 1;
    m_pWlanConfigNode->cnt     = 0;
    m_pWlanConfigNode->crypt   = 0;
    m_pWlanConfigNode->defaultConfig = 0;
    m_pWlanConfigNode->flags = 0x01; // Automatisch die Konfiguration waehlen.
    m_pWlanConfigNode->ptrWlanConfigArray = sizeof(WLANCONFIGNODE);
}

void FileWlanConfig::Init(int iCount)
{
    int i, j;

    // Pruefsumme zum Test auf Datenaenderung ruecksetzen.
    _crc = 0;

    // Speicher fuer WlanConfigDaten loeschen
    memset(m_memWlanConfig, 0x00, MAX_MEM_WLANCONFIG);

    // 1.
    // WlanConfig init
    // Grundstruktur aufsetzen
    m_pWlanConfigNode->version = 1;
    m_pWlanConfigNode->cnt     = iCount;
    m_pWlanConfigNode->crypt   = 0;
    m_pWlanConfigNode->defaultConfig = 0;
    m_pWlanConfigNode->flags   = 0x01; // Automatisch die Konfiguration waehlen.
    m_pWlanConfigNode->ptrWlanConfigArray = sizeof(WLANCONFIGNODE);

    // Zeigerarray und WlanConfigstrukturzeiger initialisieren
    i = m_pWlanConfigNode->cnt;
    for (j=0; j < i; j++) {

        // Zeigereintrag auf WLAN Konfiguration
        *(unsigned int*)&m_memWlanConfig[m_pWlanConfigNode->ptrWlanConfigArray + j * sizeof(unsigned int)] = 
            m_pWlanConfigNode->ptrWlanConfigArray + i * sizeof(unsigned int) + j * sizeof(WLANCONFIG);

        m_pWlanConfig[j] = (WLANCONFIG*)&m_memWlanConfig[m_pWlanConfigNode->ptrWlanConfigArray + i * sizeof(unsigned int) + j * sizeof(WLANCONFIG)];
        DefaultWlanConfig(*m_pWlanConfig[j]);
    }

    // Variable fuer Groessenangabe
    m_sizeData = sizeof(WLANCONFIGNODE) + m_pWlanConfigNode->cnt * sizeof(unsigned int) + m_pWlanConfigNode->cnt * sizeof(WLANCONFIG);
    if (m_sizeData > MAX_MEM_WLANCONFIG) {
      
        throw 0;
    }
}

int FileWlanConfig::SetWlanConfig(int idx, const WLANCONFIG &wc)
{
    if (idx >= m_pWlanConfigNode->cnt)
    {
        return -1;
    }

    *m_pWlanConfig[idx] = wc;
    return idx;
}

WLANCONFIG FileWlanConfig::GetWlanConfig(int idx)
{
    WLANCONFIG wc;
    memset(&wc, 0, sizeof(WLANCONFIG));

    if (idx >= m_pWlanConfigNode->cnt)
    {
        return wc;
    }

    return *m_pWlanConfig[idx];
}

void FileWlanConfig::WlanConfigCRC(int idx)
{
    int i;
    if (idx == -1) { // Alle Verfuegbaren.

        idx = GetWlanConfigCount();
        for (i = 0; i < idx; i++) {

            m_pWlanConfig[i]->crc = GetCRC16((unsigned char*)m_pWlanConfig[i], sizeof(WLANCONFIG) - sizeof(unsigned short));
        }
    } else if (idx >= 0 && idx < GetWlanConfigCount()){

        m_pWlanConfig[idx]->crc = GetCRC16((unsigned char*)m_pWlanConfig[idx], sizeof(WLANCONFIG) - sizeof(unsigned short));
    }
}

int FileWlanConfig::WlanConfigCheckCRC(int idx)
{
    int i;
    if (idx == -1) { // Alle Verfuegbaren.

        idx = GetWlanConfigCount();
        for (i = 0; i < idx; i++) {

            if (m_pWlanConfig[i]->crc != GetCRC16((unsigned char*)m_pWlanConfig[i], sizeof(WLANCONFIG) - sizeof(unsigned short)))
                return 0;
        }
    } else if (idx >= 0 && idx < GetWlanConfigCount()){

        if (m_pWlanConfig[idx]->crc != GetCRC16((unsigned char*)m_pWlanConfig[idx], sizeof(WLANCONFIG) - sizeof(unsigned short)))
            return 0;
    }

    return 1;
}

void FileWlanConfig::WlanConfigCrypt(int idx)
{
    // !!! ACHTUNG: Routine verwendet zuvor zu setzenden Schluessel !!!
    int i;
    if (idx == -1) { // Alle Verfuegbaren.

        idx = GetWlanConfigCount();
        for (i = 0; i < idx; i++) {

            _crypt.Init();
            _crypt.Crypt((unsigned char*)m_pWlanConfig[i], sizeof(WLANCONFIG));
        }
    } else if (idx >= 0 && idx < GetWlanConfigCount()) {

        _crypt.Init();
        _crypt.Crypt((unsigned char*)m_pWlanConfig[idx], sizeof(WLANCONFIG));
    }
}

int FileWlanConfig::Load(const char *fileName)
{
    // Lesen der Hexdatei
    FILE *fp;
    char data[0x400];

    // Fehlerobjekt zuruecksetzen
    resetError();

    // Umgebung zuruecksetzen
    reset();

    if ((fp = fopen(fileName, "r")) == (FILE*)0) {
        // Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    // Dateinamen halten
    strncpy(_fileName, fileName, sizeof(_fileName) - 1);
    // Pruefsumme zum Test auf Datenaenderung loeschen.
    _crc = 0;


    // Daten eintragen
    unsigned char type;
    unsigned short address;

    while(!feof(fp) && (errorState() == None)) {

        if (fgets(data, 0x400, fp) != 0) {

            address = GetAddress(data);
            type    = GetType(data);

            switch (type) {

			case TypeWlanConfigData:
                m_sizeData += GetData(data, &m_memWlanConfig[address]);
                break;

			case TypeEOF:
                break;
            }
        }
    }
    fclose(fp);	

    // Member durch gelesene Daten initialisieren und pruefen
    if (m_sizeData) {

        int i, j;

        // 1.
        // WLAN-Konfiguration pruefen

        // Zeigerarray und Wlanstrukturzeiger initialisieren
        i = m_pWlanConfigNode->cnt;
        if (i > MAX_WLANCONFIG) {
			// Fehler zuweisen
			setErrorState( Format );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }

        for (j=0; j < i; j++) {

            m_pWlanConfig[j] = (WLANCONFIG*)&m_memWlanConfig[m_pWlanConfigNode->ptrWlanConfigArray + i * sizeof(unsigned int) + j * sizeof(WLANCONFIG)];
        }

        // Pruefsumme zum Test auf Datenaenderung berechnen.
        _crc = GetCRC32(m_memWlanConfig, MAX_MEM_WLANCONFIG);
    } else {

		// Fehler zuweisen
		setErrorState( NoData );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    return 1;
}

int FileWlanConfig::Save(const char *fileName, unsigned char append /*= 0*/)
{
    FILE *fp;
    char szData[0x400];

    // Fehlerobjekt zuruecksetzen
    resetError();

    // vorhandene Datei anlegen
    if (!append) {
        if ((fp = fopen(fileName, "w")) == (FILE*)0) {
            // Oeffnen der Datei gescheitert
			setErrorState( FileOpen );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }
        fclose(fp);
    }

 
    // Daten der WLAN-Konfiguration sichern.
    if (!SaveWlanConfig(fileName, 1)) {

        return 0;
    }

    if (!append) {
        // Endezeile in Datei eintragen
        if ((fp = fopen(fileName, "a")) == (FILE*)0) {
            // Oeffnen der Datei gescheitert
			setErrorState( FileOpen );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }

        if (!MakeEndLine(szData)) {

            fclose(fp);
			// Fehler zuweisen
			setErrorState( RecordEndMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }

        fprintf(fp, "%s\n", szData);
        fclose(fp);
    }

    // Dateinamen halten
    strncpy(_fileName, fileName, sizeof(_fileName) - 1);
    // Pruefsumme zum Test auf Datenaenderung berechnen.
    _crc = GetCRC32(m_memWlanConfig, MAX_MEM_WLANCONFIG);

    return 1;
}

int FileWlanConfig::SaveWlanConfig(const char *fileName, unsigned char append /*= 0*/)
{
    FILE *fp;
    char szData[0x400];

    if ((fp = fopen(fileName, (append) ? "a" : "w")) == (FILE*)0) {
        // Oeffnen der Datei gescheitert
		setErrorState( FileOpen );
		setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
        return 0;
    }

    // Daten der WLAN-Konfiguration eintragen
    unsigned int ulPos;
    for (ulPos=0; ulPos < m_sizeData; ulPos += maxDataSize()) {

        if (!MakeTypedDataLine((unsigned short)ulPos, TypeWlanConfigData, 
            &m_memWlanConfig[ulPos], (unsigned char)(((ulPos + maxDataSize()) < m_sizeData) ? maxDataSize() : m_sizeData - ulPos), szData)) {

				// Fehler zuweisen
				setErrorState( RecordMake );
				setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
                break;
        }

        fprintf(fp, "%s\n", szData);
    }

    // Abbruch nach Fehler
    if (errorState() != None) {

        fclose(fp);
        return 0;
    }

    if (!append) {

        if (!MakeEndLine(szData)) {

            fclose(fp);
			// Fehler zuweisen
			setErrorState( RecordEndMake );
			setErrorDetail( boost::str( boost::format( "%1%(%2%): %3%" ) % __FILE__ % __LINE__ % errorText() ) );
            return 0;
        }

        fprintf(fp, "%s\n", szData);
    }

    fclose(fp);

    return 1;
}

int FileWlanConfig::DefaultWlanConfig(WLANCONFIG &wc)
{
    memset(&wc, 0, sizeof(WLANCONFIG));

    //wc.SSID[33];
    wc.EnableWLAN               = 1;
    wc.Type                     = 0;
    wc.Band                     = 0;
    wc.Channel                  = 0;
    wc.Country                  = 3;
    wc.Security                 = 0;
    wc.Authentication           = 0;
    wc.PairwiseEncryptionMethod = 0;
    wc.GroupEncryptionMethod    = 0;
    //wc.Key[64];
    wc.KeyType                  = 0;
    wc.TxKeyIndex               = 0;
    wc.Autofallback             = 1;
    wc.TxDataRate               = 0;
    wc.RadioPower               = 0;
    wc.PowerLevel               = 2;
    wc.DHCP                     = 1;
    strcpy((char*)wc.IpAddress, "0.0.0.0");
    wc.Port                     = 8000;
    strcpy((char*)wc.Hostbits, "008");
    strcpy((char*)wc.SubNetWorkMask, "255.255.255.0");
    strcpy((char*)wc.GateWay, "0.0.0.0");
    wc.RemoteAccess             = 1;
    wc.Alive                    = 45;
    //wc.Name[33];
    wc.timeout                  = 60;
    wc.flags                    = 0;
    wc.startupTimeout           = 15;

    return 1;
}

